
from typing import Optional, Tuple, Any, List, Dict, Union
from collections import OrderedDict
from returnn.tf.layers.basic import LayerBase
from .module import Module
from ...tensor import Tensor, dtype as _dtype
from ....naming import Naming, TensorEntry


class Copy(Module):
  is_original_torch_module = False

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    return {"class": "copy", "from": self._get_input_layer_name(input)}

  def make_output_tensor_from_returnn(self, inputs_flat: List[Tensor], layer: LayerBase) -> Tensor:
    assert len(inputs_flat) == 1
    return inputs_flat[0]


class Range(Module):
  is_original_torch_module = False

  def create_returnn_layer_dict(self, limit, start, delta, dtype, sparse=False) -> Dict[str, Any]:
    return {"class": "range", "limit": limit, "start": start, "delta": delta, "dtype": dtype, "sparse": sparse}


class Cat(Module):
  is_original_torch_module = False

  def __init__(self, dim=0):
    super(Cat, self).__init__()
    self.dim = dim

  def create_returnn_layer_dict(self, *inputs: Tensor) -> Dict[str, Any]:
    naming = Naming.get_instance()
    sources = []
    for input in inputs:
      dim = self.dim
      if dim < 0:
        dim += len(input.shape)
      assert 0 <= dim < len(input.shape)
      returnn_axis = self._get_input_axis_to_returnn(input, axis=dim)
      sources.append((self._get_input_layer_name(input), returnn_axis))
    return {"class": "concat", "from": sources}


class GetSublayer(Module):
  is_original_torch_module = False

  def __init__(self, sub_layer: str):
    super(GetSublayer, self).__init__()
    self.sub_layer = sub_layer

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    return {"class": "copy", "from": f"{self._get_input_layer_name(input)}/{self.sub_layer}"}


class Cast(Module):
  is_original_torch_module = False

  def __init__(self, dtype: Union[str, _dtype]):
    super(Cast, self).__init__()
    self.dtype = _dtype(dtype)

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    return {"class": "cast", "from": self._get_input_layer_name(input), "dtype": self.dtype.name}


class BinaryOperator(Module):
  is_original_torch_module = False

  def __init__(self, kind: str):
    """
    :param kind: "add", "sub", "mul", "truediv"
    """
    super(BinaryOperator, self).__init__()
    self.kind = kind

  def get_returnn_name(self) -> str:
    return self.kind

  def create_returnn_layer_dict(self, *inputs: Tensor):
    inputs = _unify_tensor_axes_returnn_meta(*inputs)
    return {
      "class": "combine", "kind": self.kind,
      "from": [self._get_input_layer_name(input) for input in inputs]}


class ComparisonOperator(BinaryOperator):
  is_original_torch_module = False

  def __init__(self, kind: str):
    """
    :param str kind: e.g. "equal", "greater", "less" or other supported TF comparison ops
    """
    super(ComparisonOperator, self).__init__(kind=kind)

  def create_returnn_layer_dict(self, *inputs: Tensor):
    inputs = _unify_tensor_axes_returnn_meta(*inputs)
    return {
      "class": "compare", "kind": self.kind,
      "from": [self._get_input_layer_name(input) for input in inputs]}


class Reciprocal(Module):
  """
  1/x or 1/max(eps,x)
  """
  is_original_torch_module = False

  def __init__(self, eps: Optional[float] = None):
    super(Reciprocal, self).__init__()
    self.eps = eps

  def create_returnn_layer_dict(self, input: Tensor):
    x = "source(0)"
    if self.eps is not None:
      x = f"maximum_with_identity_grad({x})"
    return {
      "class": "eval", "eval": f"tf_compat.v1.reciprocal({x})",
      "from": self._get_input_layer_name(input)}


class Max(Module):
  is_original_torch_module = False
  pass  # TODO


class ReturnnReinterpretSameSizeAs(Module):
  """
  Used by :func:`_unify_tensor_axes_returnn_meta`.
  """
  is_original_torch_module = False

  def create_returnn_layer_dict(self, input: Tensor, same_as: Tensor) -> Dict[str, Any]:
    return {
      "class": "reinterpret_data",
      "from": self._get_input_layer_name(input),
      "size_base": self._get_input_layer_name(same_as)}

  def _get_output_shape_from_returnn(self,
                                     inputs_flat: List[Tensor], layer: LayerBase
                                     ) -> Tuple[Tuple[int, ...], Dict[int, int]]:
    input, _ = inputs_flat
    assert isinstance(input, Tensor)
    naming = Naming.get_instance()
    x = naming.tensors[input]
    assert isinstance(x, TensorEntry)
    return input.shape, x.returnn_axis_from_torch_axis  # no change


class ReturnnReinterpretSetAxes(Module):
  """
  Used by :func:`_unify_tensor_axes_returnn_meta`.
  """
  is_original_torch_module = False

  def __init__(self, *,
               dims_by_key: Optional[Dict[str, int]] = None,  # "B"|"T"|"F" are the keys
               batch_dim: Optional[int] = None,
               time_dim: Optional[int] = None,
               feature_dim: Optional[int] = None):
    super(ReturnnReinterpretSetAxes, self).__init__()
    self.dims_by_key = dims_by_key
    self.batch_dim = batch_dim
    self.time_dim = time_dim
    self.feature_dim = feature_dim

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    axes = {}
    if self.dims_by_key:
      for key, dim in self.dims_by_key.items():
        axes[key] = self._get_input_axis_to_returnn(input, dim)
    for key, dim in [("B", self.batch_dim), ("T", self.time_dim), ("F", self.feature_dim)]:
      if dim is None:
        continue
      axes[key] = self._get_input_axis_to_returnn(input, dim)
    return {
      "class": "reinterpret_data",
      "from": self._get_input_layer_name(input),
      "set_axes": axes}

  def _get_output_shape_from_returnn(self,
                                     inputs_flat: List[Tensor], layer: LayerBase
                                     ) -> Tuple[Tuple[int, ...], Dict[int, int]]:
    input, = inputs_flat
    assert isinstance(input, Tensor)
    naming = Naming.get_instance()
    x = naming.tensors[input]
    assert isinstance(x, TensorEntry)
    return input.shape, x.returnn_axis_from_torch_axis  # no change


class Transpose(Module):
  """
  Note: The resulting Torch tensor is transposed as expected.
  However, on the RETURNN side, we actually should never need to transpose,
  as we have dimension tags, and all layers should refer to axes by dim tags.
  So on RETURNN side, this is a no-op.
  """
  is_original_torch_module = False

  def __init__(self, perm: Optional[Union[Dict[int, int], Tuple[int, ...], List[int]]]):
    super(Transpose, self).__init__()
    if perm is None:
      self.perm = None
    else:
      if isinstance(perm, (list, tuple)):
        perm = dict(enumerate(perm))
      assert isinstance(perm, dict)
      self.perm = perm  # type: Optional[Dict[int, int]]
      assert len(perm) == len(set(perm.values()))  # should be unique

  def _get_perm(self, input: Tensor) -> Dict[int, int]:
    if self.perm is None:
      dims = range(len(input.shape))
      return {i: j for (i, j) in zip(dims, reversed(dims))}

    def _axis(i):
      assert -len(input.shape) <= i < len(input.shape)
      if i < 0:
        i += len(input.shape)
      assert 0 <= i < len(input.shape)
      return i
    perm = {_axis(i): _axis(j) for (i, j) in self.perm.items()}
    assert len(perm) == len(set(perm.values()))  # should be unique
    return perm

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    # See comment in class docstring. No need to actually transpose.
    return {"class": "copy", "from": self._get_input_layer_name(input)}

  def _get_output_shape_from_returnn(self,
                                     inputs_flat: List[Tensor], layer: LayerBase
                                     ) -> Tuple[Tuple[int, ...], Dict[int, int]]:
    """
    :return: (torch_shape, returnn_axis_from_torch_axis).
      Torch shape how it would have looked when this would be processed within Torch.
      The RETURNN layer.output shape (order of axes) might look different.

    On RETURNN side, this is a no-op.
    But on Torch side, we transpose as expected.
    """
    input, = inputs_flat
    perm = self._get_perm(input)
    target_axes = set(perm)
    source_axes = set(perm.values())
    ndim = len(input.shape)
    rem_target_axes = [i for i in range(ndim) if i not in target_axes]
    rem_source_axes = [i for i in range(ndim) if i not in source_axes]
    assert len(rem_target_axes) == len(rem_source_axes)
    perm.update({i: j for (i, j) in zip(rem_target_axes, rem_source_axes)})
    assert len(perm) == len(set(perm.values())) == ndim

    naming = Naming.get_instance()
    tensor_entry = naming.tensors[input]
    assert isinstance(tensor_entry, TensorEntry)
    assert tensor_entry.returnn_data and tensor_entry.returnn_axis_from_torch_axis
    assert tensor_entry.returnn_data.batch_ndim == ndim

    inv_perm = {j: i for i, j in perm.items()}
    out_torch_shape = [input.shape[perm[i]] for i in range(ndim)]
    out_returnn_axis_from_torch_axis = {
      inv_perm[i]: j for (i, j) in tensor_entry.returnn_axis_from_torch_axis.items()}
    return tuple(out_torch_shape), out_returnn_axis_from_torch_axis


class Gather(Module):
  """
  Basically x[pos] but in specific dim (axis).
  """
  is_original_torch_module = False

  def __init__(self, dim: int, pos: int):
    super(Gather, self).__init__()
    self.dim = dim
    self.pos = pos

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    return {
      "class": "gather", "from": self._get_input_layer_name(input),
      "axis": self._get_input_axis_to_returnn(input, axis=self.dim),
      "position": self.pos}


class Slice(Module):
  """
  Slicing of tensors. Wraps RETURNN's SliceLayer.
  """
  is_original_torch_module = False

  def __init__(self, axis: int, start: Optional[int] = None, stop: Optional[int] = None, step: Optional[int] = None):
    super(Slice, self).__init__()
    self.axis = axis
    self.start = start
    self.stop = stop
    self.step = step

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    return {
      "class": "slice", "axis": self._get_input_axis_to_returnn(input, axis=self.axis),
      "slice_start": self.start, "slice_end": self.stop, "slice_step": self.step,
      "from": self._get_input_layer_name(input)}

  def _get_output_shape_from_returnn(self, inputs_flat: List[Tensor], layer: LayerBase
                                     ) -> Tuple[Tuple[int, ...], Dict[int, int]]:
    """
    The size of the dynamic axes might be changed, so we have to take care of this here for the torch shape.
    """
    torch_shape, returnn_axis_from_torch_axis = super(Slice, self)._get_output_shape_from_returnn(
      inputs_flat=inputs_flat, layer=layer)
    assert len(inputs_flat) == 1
    torch_shape = list(inputs_flat[0].shape)
    start = self.start or 0
    stop = self.stop or torch_shape[self.axis]
    step = self.step or 1
    if start < 0:
      start += torch_shape[self.axis]
    if stop < 0:
      stop += torch_shape[self.axis]
    assert 0 <= start <= torch_shape[self.axis] - 1
    assert 0 <= stop <= torch_shape[self.axis]
    torch_shape[self.axis] = len(range(start, stop, step))
    return tuple(torch_shape), returnn_axis_from_torch_axis


class Stack(Module):
  """
  Wraps RETURNN StackLayer.
  """
  is_original_torch_module = False

  def __init__(self, dim: Optional[int] = None):
    super(Stack, self).__init__()
    self.dim = dim

  def create_returnn_layer_dict(self, *inputs: Tensor) -> Dict[str, Any]:
    return {
      "class": "stack", "axis": self.dim,
      "from": [self._get_input_layer_name(x) for x in inputs]}


class Tile(Module):
  """
  Wraps RETURNN TileLayer.
  """
  is_original_torch_module = False

  def __init__(self, multiples: Dict[int, int]):
    super(Tile, self).__init__()
    assert isinstance(multiples, dict)
    self.multiples = multiples

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    multiples = {
      self._get_input_axis_to_returnn(input, axis=axis): multiple
      for axis, multiple in self.multiples.items()}
    return {
      "class": "tile", "multiples": multiples,
      "from": self._get_input_layer_name(input)}


class Reduce(Module):
  """
  Wraps RETURNN ReduceLayer.
  """
  is_original_torch_module = False

  def __init__(self, mode: str, axes: Union[str, int]):
    assert mode in ["sum", "max", "argmin", "min", "argmax", "mean", "logsumexp"]
    super(Reduce, self).__init__()
    self.mode = mode
    self.axes = axes

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    return {"class": "reduce", "mode": self.mode, "axes": self.axes, "from": self._get_input_layer_name(input)}


def _unify_tensor_axes_returnn_meta(*inputs: Tensor) -> Tuple[Tensor, ...]:
  """
  You have multiple inputs which can potentially have different dynamic axes (see RETURNN :class:`Data`),
  and this would add ``reinterpret_data`` layers when needed
  to make sure the seq lengths / dim tags are the same.

  It also makes sure that feature_dim_axis and time_dim_axis (the RETURNN flags) are set consistently.

  It also makes sure that broadcasting in the batch-dim axis works correctly
  by potentially removing such a batch broadcast dim.
  """
  if len(inputs) <= 1:
    return inputs
  naming = Naming.get_instance()
  tensors = [naming.tensors[x] for x in inputs if isinstance(x, Tensor)]  # type: List[TensorEntry]
  tensors = [x for x in tensors if x.returnn_data.batch_ndim > 0]  # filter out scalars
  if len(tensors) <= 1:
    return inputs
  inputs = list(inputs)

  dims = {}  # torch axis -> (tensor, DimensionTag) (static != 1, or dynamic)
  for i, x in enumerate(inputs):
    if not isinstance(x, Tensor):
      continue
    x = naming.tensors[x]
    assert isinstance(x, TensorEntry)
    if x.returnn_data.batch_ndim == 0:  # scalars are fine
      continue
    used_reinterpret_same_size = False
    for axis in range(-1, -x.returnn_data.batch_ndim - 1, -1):
      returnn_axis = x.returnn_axis_from_torch_axis[axis + x.returnn_data.batch_ndim]
      dim_tag = x.returnn_data.get_dim_tag(returnn_axis)
      if dim_tag.dimension == 1:
        continue  # broadcast dim, so always ok. do not add
      if axis not in dims:
        dims[axis] = (x, dim_tag)
      else:
        prev_x, prev_dim_tag = dims[axis]
        if prev_dim_tag is dim_tag:
          pass  # ok
        elif prev_dim_tag.dimension is not None:
          # Static dimension, need to match.
          assert dim_tag.dimension == prev_dim_tag.dimension, f"not matching Data {x.returnn_data}"
        else:
          assert not used_reinterpret_same_size  # It should not happen to use this multiple times.
          # This is the case we wanted to catch here.
          # We have two different dynamic axes, which are expected to represent the same seq lengths.
          # So we unify them via reinterpret_data.
          x_ = ReturnnReinterpretSameSizeAs()(x.tensor(), prev_x.tensor())
          x = naming.tensors[x_]
          assert isinstance(x, TensorEntry)
          inputs[i] = x_
          used_reinterpret_same_size = True

  # Figure out in which axis we have the batch/time/feature axis.
  special_axes_names = [("B", "batch_dim_axis"), ("T", "time_dim_axis"), ("F", "feature_dim_axis")]
  torch_special_axes = {}  # "B"|"T"|"F" -> axis->dict
  for i, x in enumerate(inputs):
    if not isinstance(x, Tensor):
      continue
    x = naming.tensors[x]
    assert isinstance(x, TensorEntry)
    for key, axis_name in special_axes_names:
      returnn_axis = getattr(x.returnn_data, axis_name)
      if returnn_axis is None:
        continue
      returnn_dim = x.returnn_data.batch_shape[returnn_axis]
      torch_axis = x.torch_axis_from_returnn_axis[returnn_axis] - x.returnn_data.batch_ndim
      if key not in torch_special_axes:
        torch_special_axes[key] = OrderedDict({torch_axis: returnn_dim})
      elif torch_axis not in torch_special_axes[key] or torch_special_axes[key][torch_axis] == 1:
        assert key != "B", f"mismatch in batch axis in {inputs!r}"
        torch_special_axes[key][torch_axis] = returnn_dim

  axes_remap_per_input = {}  # input idx -> remap dict
  for key, axis_name in special_axes_names:
    if key not in torch_special_axes:
      continue
    torch_axes = torch_special_axes[key]
    if len(torch_axes) <= 1:
      continue
    assert key != "B"  # should be caught already...
    if any(d != 1 for _, d in torch_axes.items()):
      torch_axes = {axis: dim for (axis, dim) in torch_axes.items() if dim != 1}  # filter out broadcast specs
    torch_ref_axis, torch_ref_dim = list(torch_axes.items())[0]
    assert all(torch_ref_dim == d for _, d in torch_axes.items())
    for i, x in enumerate(inputs):
      if not isinstance(x, Tensor):
        continue
      x = naming.tensors[x]
      assert isinstance(x, TensorEntry)
      if x.returnn_data.batch_ndim == 0:  # ignore scalars
        continue
      returnn_axis = getattr(x.returnn_data, axis_name)
      if returnn_axis is None or x.torch_axis_from_returnn_axis[returnn_axis] != torch_ref_axis:
        axes_remap_per_input.setdefault(i, {})[key] = torch_ref_axis
  for i, axes_remap in axes_remap_per_input.items():
    x = inputs[i]
    assert isinstance(x, Tensor)
    inputs[i] = ReturnnReinterpretSetAxes(dims_by_key=axes_remap)(x)

  if "B" in torch_special_axes:
    torch_batch_axes = torch_special_axes["B"]
    assert len(torch_batch_axes) == 1
    torch_batch_axis = list(torch_batch_axes)[0]
    # Now (lastly) get rid of potential broadcast batch dims.
    from .shape import Squeeze
    for i, x in enumerate(inputs):
      if not isinstance(x, Tensor):
        continue
      x = naming.tensors[x]
      assert isinstance(x, TensorEntry)
      if x.returnn_data.batch_ndim == 0:  # ignore scalars
        continue
      if not x.returnn_data.have_batch_axis():
        returnn_batch_axis = x.returnn_axis_from_torch_axis.get(torch_batch_axis + x.returnn_data.batch_ndim, None)
        if returnn_batch_axis is not None:
          assert x.returnn_data.batch_shape[returnn_batch_axis] == 1  # expected to be broadcast dim
          # In RETURNN, that's not needed, and that also must not be done like that,
          # i.e. broadcasting in the batch dim can only be implicit (i.e. by missing batch-dim).
          # Note that now the number of dimensions do not match anymore!
          # But this should not matter for the use cases of this function.
          inputs[i] = Squeeze(dim=torch_batch_axis)(x.tensor())

  return tuple(inputs)


__all__ = [
  key for (key, value) in sorted(globals().items())
  if not key.startswith("_")
  and getattr(value, "__module__", "") == __name__]



"""
Code from here:
https://github.com/lucidrains/conformer/blob/master/conformer/conformer.py
"""

import torch
from torch import nn
import torch.nn.functional as F


# helper functions

def calc_same_padding(kernel_size):
    pad = kernel_size // 2
    return (pad, pad - (kernel_size + 1) % 2)


# helper classes

class Swish(nn.Module):
    def forward(self, x):
        return x * x.sigmoid()


class Transpose(nn.Module):
    def __init__(self, dims):
        super().__init__()
        assert len(dims) == 2, 'dims must be a tuple of two dimensions'
        self.dims = dims

    def forward(self, x):
        return x.transpose(*self.dims)


class GLU(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.dim = dim

    def forward(self, x):
        out, gate = x.chunk(2, dim=self.dim)
        return out * gate.sigmoid()


class DepthWiseConv1d(nn.Module):
    def __init__(self, chan_in, chan_out, kernel_size, padding):
        super().__init__()
        self.padding = padding
        self.conv = nn.Conv1d(chan_in, chan_out, kernel_size, groups = chan_in)

    def forward(self, x):
        x = F.pad(x, self.padding)
        return self.conv(x)


# main class

class ConformerConvModule(nn.Module):
    def __init__(
        self,
        dim,
        causal = False,
        expansion_factor = 2,
        kernel_size = 31,
        dropout = 0.):
        super().__init__()

        inner_dim = dim * expansion_factor
        padding = calc_same_padding(kernel_size) if not causal else (kernel_size - 1, 0)

        self.net = nn.Sequential(
            nn.LayerNorm(dim),
            Transpose((1, 2)),
            nn.Conv1d(dim, inner_dim * 2, 1),
            GLU(dim=1),
            DepthWiseConv1d(inner_dim, inner_dim, kernel_size = kernel_size, padding = padding),
            nn.BatchNorm1d(inner_dim) if not causal else nn.Identity(),
            Swish(),
            nn.Conv1d(inner_dim, dim, 1),
            Transpose((1, 2)),
            nn.Dropout(dropout)
        )

    def forward(self, x):
        return self.net(x)



#!rnn.py

config = globals()["config"]  # make IDEs happy

# mode selection
MODE = config.value("mode", "train")
assert MODE in ['train', 'gta_decoding']

# global settings
debug_mode = False
debug_print_layer_output_shape=False
debug_print_layer_output_template=True
device = 'gpu'
log = ['./returnn.log']
log_batch_size = True
log_verbosity = 5
# model = 'returnn_models/window_epoch_t2'
model = 'returnn_models/window_epoch'

# choose eval_task if we run GTA decoding
if MODE=='gta_decoding':
    task='eval'
else:
    task = 'train'

# tensorflow settings
tf_log_memory_usage = True

# learning rate and gradient control
accum_grad_multiple_step = 2
gradient_clip = 1
gradient_noise = 0
learning_rate_control = 'newbob_multi_epoch'
learning_rate_control_min_num_epochs_per_new_lr = 5
learning_rate_control_relative_error_relative_lr = True
learning_rate_file = 'returnn_training_lr_scores'
learning_rates = [0.001]
use_learning_rate_control_always = True

# newbob learning rate control
newbob_learning_rate_decay = 0.9
newbob_multi_num_epochs = 5
newbob_multi_update_interval = 1
newbob_relative_error_threshold = 0

# gradient optimizer
optimizer = {'class': 'adam', 'epsilon': 1e-08}

# data feeding
batch_size = 28000
extern_data = {
    'classes': {'dim': 77, 'shape': (None,), 'sparse': True, 'available_for_inference': True},
    'data': {'dim': 80, 'shape': (None, 80), 'available_for_inference': False},
}
max_seq_length = {'data': 1000}
max_seqs = 200

#training
cleanup_old_models = True
num_epochs = 200
save_interval = 1
stop_on_nonfinite_train_score = False

# dataset defintion
def get_dataset(type):
    global_parameters = { 
        'class': 'OggZipDataset',
        'audio': { 
            'feature_options': {'fmin': 60, 'fmax': 7600},
            'features': 'db_mel_filterbank',
            'join_frames': 1,
            'norm_mean': -74.8469952583313,
            'norm_std_dev': 32.351323967712815,
            'num_feature_filters': 80,
            'peak_normalization': False,
            'preemphasis': 0.97,
            'step_len': 0.0125,
            'window_len': 0.05},
        'targets': {
            'class': 'Vocabulary',
            'unknown_label': None,
            'vocab_file': 'data/cmu_vocab.pkl'},
        'path': 'data/LJSpeech.ogg.zip',
        'partition_epoch': 1,
    }
    if type == "dev":
        dev_parameters = {
            'segment_file': 'data/dev_segments.txt',
            'seq_ordering': 'sorted',
        }
        dataset_parameters = {**global_parameters, **dev_parameters}
    if type == "train":
        train_parameters = {
            'segment_file': 'data/train_segments.txt',
            'seq_ordering': 'laplace:.1000',
        }
        dataset_parameters = {**global_parameters, **train_parameters}
    return dataset_parameters

if MODE=='gta_decoding':
    dev = get_dataset("train")
    train = None
else:
    dev = get_dataset("dev")
    train = get_dataset("train")

# Network template
network_template = { 
    'decoder': { 'cheating': False,
        'class': 'rec',
        'from': [],
        'max_seq_len': 1000,
        'target': 'windowed_data_target',
        'unit': { 
            'accum_att_weights': {'class': 'combine',
                                  'kind': 'add',
                                  'from': ['prev:accum_att_weights', 'att_weights'],
                                  'is_output_layer': True},
            'att0': {'base': 'base:encoder', 'class': 'generic_attention', 'weights': 'att_weights'},
            'att_energy': {'activation': None, 'class': 'linear', 'from': ['att_energy_tanh'], 'n_out': 1, 'with_bias': False},
            'att_energy_in': { 'class': 'combine',
                               'from': ['base:enc_ctx', 's_transformed', 'location_feedback_transformed'],
                               'kind': 'add',
                               'n_out': 128},
            'att_energy_tanh': {'activation': 'tanh', 'class': 'activation', 'from': ['att_energy_in']},
            'att_weights': {'class': 'softmax_over_spatial', 'from': ['att_energy']},
            'choice': {'beam_size': 1, 'class': 'choice', 'from': ['output'],
                       'input_type': 'regression', 'target': 'windowed_data_target',
                       'n_out': 160},
            'convolved_att': { 'L2': 1e-07,
                               'activation': None,
                               'class': 'conv',
                               'filter_size': (31,),
                               'from': ['feedback_pad_right'],
                               'n_out': 32,
                               'padding': 'valid'},
            'decoder_1': { 'class': 'rnn_cell',
                           'from': ['pre_net_layer_2_out', 'prev:att0'],
                           'n_out': 640,
                           'unit': 'zoneoutlstm',
                           'unit_opts': {'zoneout_factor_cell': 0.1, 'zoneout_factor_output': 0.1}},
            'decoder_2': { 'class': 'rnn_cell',
                           'from': ['decoder_1'],
                           'n_out': 640,
                           'unit': 'zoneoutlstm',
                           'unit_opts': {'zoneout_factor_cell': 0.1, 'zoneout_factor_output': 0.1}},
            'end_compare': {'class': 'compare', 'from': ['stop_token_sigmoid'], 'kind': 'greater', 'value': 0.5},
            'end': {'class': 'squeeze', 'from': ['end_compare'], 'axis': 'F'},
            'entropy': { 'class': 'eval',
                         'eval': '-tf.reduce_sum(source(0)*safe_log(source(0)), axis=-1, keepdims=True)',
                         'from': ['att_weights'],
                         'loss': 'as_is',
                         'loss_scale': 0.0001},
            'feedback_pad_left': { 'axes': 's:0',
                                   'class': 'pad',
                                   'from': ['prev:accum_att_weights'],
                                   'mode': 'constant',
                                   'padding': ((15, 0),),
                                   'value': 1},
            'feedback_pad_right': { 'axes': 's:0',
                                    'class': 'pad',
                                    'from': ['feedback_pad_left'],
                                    'mode': 'constant',
                                    'padding': ((0, 15),),
                                    'value': 0},
            'location_feedback_transformed': { 'L2': 1e-07,
                                               'activation': None,
                                               'class': 'linear',
                                               'dropout': 0.1,
                                               'from': ['convolved_att'],
                                               'n_out': 128,
                                               'with_bias': False},
            'output': { 'activation': None,
                        'class': 'linear',
                        'from': ['decoder_2', 'att0'],
                        'loss': 'mean_l1',
                        'loss_scale': 1.0,
                        'n_out': 160,
                        'target': 'windowed_data_target'},
            'pre_net_layer_1': {'L2': 1e-07, 'activation': 'relu', 'class': 'linear', 'from': ['pre_slice'], 'n_out': 128},
            'pre_net_layer_2': { 'L2': 1e-07,
                                 'activation': 'relu',
                                 'class': 'linear',
                                 'dropout': 0.5,
                                 'dropout_noise_shape': {'*': None},
                                 'dropout_on_forward': True,
                                 'from': ['pre_net_layer_1'],
                                 'n_out': 64},
            'pre_net_layer_2_out': { 'class': 'dropout',
                                     'dropout': 0.5,
                                     'dropout_noise_shape': {'*': None},
                                     'dropout_on_forward': True,
                                     'from': ['pre_net_layer_2']},
            'pre_slice': {'axis': 'F', 'class': 'slice', 'from': ['prev:choice'], 'slice_start': 80},
            's_transformed': { 'L2': 1e-07,
                               'activation': None,
                               'class': 'linear',
                               'dropout': 0.5,
                               'from': ['decoder_2'],
                               'n_out': 128,
                               'with_bias': False},
            'stop_token': { 'activation': None,
                            'class': 'linear',
                            'from': ['decoder_2', 'att0'],
                            'loss': 'bin_ce',
                            'loss_scale': 1.0,
                            'n_out': 1,
                            'target': 'stop_token_target'},
            'stop_token_sigmoid': {'activation': 'sigmoid', 'class': 'activation', 'from': ['stop_token']}}},
  'embed_batchnorm_cv_0': {'class': 'batch_norm', 'from': ['embed_conv0']},
  'embed_batchnorm_cv_1': {'class': 'batch_norm', 'from': ['embed_conv1']},
  'embed_batchnorm_cv_2': {'class': 'batch_norm', 'from': ['embed_conv2']},
  'embed_conv0': {'L2': 1e-07, 'activation': 'relu', 'class': 'conv', 'filter_size': (5,), 'from': ['embedding'], 'n_out': 128, 'padding': 'same'},
  'embed_conv0_out': {'class': 'dropout', 'dropout': 0.5, 'dropout_noise_shape': {"*": None}, 'from': ['embed_batchnorm_cv_0']},
  'embed_conv1': { 'L2': 1e-07,
                   'activation': 'relu',
                   'class': 'conv',
                   'filter_size': (5,),
                   'from': ['embed_conv0_out'],
                   'n_out': 128,
                   'padding': 'same'},
  'embed_conv1_out': {'class': 'dropout', 'dropout': 0.5, 'dropout_noise_shape': {"*": None}, 'from': ['embed_batchnorm_cv_1']},
  'embed_conv2': { 'L2': 1e-07,
                   'activation': 'relu',
                   'class': 'conv',
                   'filter_size': (5,),
                   'from': ['embed_conv1_out'],
                   'n_out': 128,
                   'padding': 'same'},
  'embed_conv2_out': {'class': 'dropout', 'dropout': 0.5, 'dropout_noise_shape': {"*": None}, 'from': ['embed_batchnorm_cv_2']},
  'embedding': {'activation': None, 'class': 'linear', 'from': ['data:classes'], 'n_out': 128},
  'enc_ctx': { 'L2': 1e-07,
               'activation': None,
               'class': 'linear',
               'dropout': 0.5,
               'from': ['encoder', 'encoder_position'],
               'n_out': 128,
               'with_bias': True},
  'encoder': {'class': 'copy', 'dropout': 0.0, 'from': ['lstm0_fw', 'lstm0_bw']},
  'encoder_position': {'class': 'positional_encoding', 'from': ['lstm0_fw'], 'n_out': 64, 'out_type': {'dim': 64, 'shape': (None, 64)}},
  'lstm0_bw': { 'class': 'rec',
                'direction': -1,
                'from': ['embed_conv2_out'],
                'n_out': 256,
                'unit': 'zoneoutlstm',
                'unit_opts': {'zoneout_factor_cell': 0.1, 'zoneout_factor_output': 0.1}},
  'lstm0_fw': { 'class': 'rec',
                'direction': 1,
                'from': ['embed_conv2_out'],
                'n_out': 256,
                'unit': 'zoneoutlstm',
                'unit_opts': {'zoneout_factor_cell': 0.1, 'zoneout_factor_output': 0.1}},
  'mse_output': {'class': 'copy', 'from': ['output'], 'loss': 'mse', 'loss_scale': 0.0, 'n_out': 80, 'target': 'padded_data_target'},
  'output': { 'class': 'combine',
              'from': ['dec_output', 'post_conv_tf'],
              'kind': 'add',
              'loss': 'mean_l1',
              'loss_scale': 0.25,
              'n_out': 80,
              'target': 'padded_data_target'},
  'dec_output_split': {'class': 'split_dims',
                   'from': ['decoder'],
                   'axis': 'F',
                   'dims': (2, -1)},
  'dec_output': {'class': 'merge_dims',
                     'from': ['dec_output_split'],
                     'axes': ['T', 'static:0'],
                     'n_out': 80,},
  'post_batchnorm_cv_0': {'class': 'batch_norm', 'from': ['post_conv0']},
  'post_batchnorm_cv_1': {'class': 'batch_norm', 'from': ['post_conv1']},
  'post_batchnorm_cv_2': {'class': 'batch_norm', 'from': ['post_conv2']},
  'post_batchnorm_cv_3': {'class': 'batch_norm', 'from': ['post_conv3']},
  'post_batchnorm_cv_4': {'class': 'batch_norm', 'from': ['post_conv4']},
  'post_conv0': { 'L2': 1e-07,
                  'activation': 'relu',
                  'class': 'conv',
                  'filter_size': (5,),
                  'from': ['dec_output'],
                  'n_out': 256,
                  'padding': 'same'},
  'post_conv0_out': {'class': 'dropout', 'dropout': 0.5, 'dropout_noise_shape': {"*": None}, 'from': ['post_batchnorm_cv_0']},
  'post_conv1': { 'L2': 1e-07,
                  'activation': 'relu',
                  'class': 'conv',
                  'filter_size': (5,),
                  'from': ['post_conv0_out'],
                  'n_out': 256,
                  'padding': 'same'},
  'post_conv1_out': {'class': 'dropout', 'dropout': 0.5, 'dropout_noise_shape': {"*": None}, 'from': ['post_batchnorm_cv_1']},
  'post_conv2': { 'L2': 1e-07,
                  'activation': 'relu',
                  'class': 'conv',
                  'filter_size': (5,),
                  'from': ['post_conv1_out'],
                  'n_out': 256,
                  'padding': 'same'},
  'post_conv2_out': {'class': 'dropout', 'dropout': 0.5, 'dropout_noise_shape': {"*": None}, 'from': ['post_batchnorm_cv_2']},
  'post_conv3': { 'L2': 1e-07,
                  'activation': 'relu',
                  'class': 'conv',
                  'filter_size': (5,),
                  'from': ['post_conv2_out'],
                  'n_out': 256,
                  'padding': 'same'},
  'post_conv3_out': {'class': 'dropout', 'dropout': 0.5, 'dropout_noise_shape': {"*": None}, 'from': ['post_batchnorm_cv_3']},
  'post_conv4': { 'L2': 1e-07,
                  'activation': 'relu',
                  'class': 'conv',
                  'filter_size': (5,),
                  'from': ['post_conv3_out'],
                  'n_out': 256,
                  'padding': 'same'},
  'post_conv4_out': {'class': 'dropout', 'dropout': 0.5, 'dropout_noise_shape': {"*": None}, 'from': ['post_batchnorm_cv_4']},
  'post_conv_tf': {'activation': None, 'class': 'conv', 'filter_size': (5,), 'from': ['post_conv4_out'], 'n_out': 80, 'padding': 'same'},
  'stop_token_target': { 'class': 'eval',
                         'eval': "self.network.get_config().typed_value('_stop_token_target')(source(0, as_data=True))",
                         'from': ['windowed_data_target'],
                         'out_type': {'dim': 1, 'shape': (None, 1)},
                         'register_as_extern_data': 'stop_token_target'},
  'windowed_data': {'class': 'window',
                    'from': ['data'],
                    'window_size': 2,
                    'window_right': 1,
                    'stride': 2
                    },
  'padded_data_target': {'class': 'merge_dims',
                         'from': ['windowed_data'],
                         'axes': ['T', 'static:0'],
                         'n_out': 80,
                         'register_as_extern_data': 'padded_data_target',},
  'windowed_data_target': {'class': 'merge_dims',
                           'from': ['windowed_data'],
                           'axes': 'static',
                           'n_out': 160,
                           'register_as_extern_data': 'windowed_data_target',}
  }


if MODE=='gta_decoding':
    network_template['dump_output'] = {'class': 'hdf_dump', 'filename': 'data/gta_decode.hdf', 'from': 'output', 'is_output_layer': True}

# Dynamic network construction
def get_network(epoch, **kwargs):
    pretrain_stage = epoch // 5
    network = network_template
    if pretrain_stage < 5:
        postnet_loss_scale = max(min((pretrain_stage/5*0.25), 0.25), 0.01)
        stop_token_loss_scale = min(pretrain_stage/5, 1.0) 
        network['output']['loss_scale'] = postnet_loss_scale
        network['decoder']['unit']['stop_token']['loss_scale'] = stop_token_loss_scale 

    return network


def _stop_token_target(data, ramp_length=5):
  import tensorflow as tf
  time_axis = data.get_dynamic_axes()[0]
  stop_position = tf.expand_dims(data.get_dynamic_size(time_axis), axis=1) - ramp_length - 1
  ramp = tf.expand_dims(tf.range(tf.shape(data.placeholder)[1]), axis=0)
  full_ramp = tf.tile(ramp, [tf.shape(data.placeholder)[0], 1])
  adapted_ramp = tf.minimum(tf.maximum(full_ramp - stop_position, 0), ramp_length)
  return tf.cast(tf.expand_dims(adapted_ramp, 2), dtype="float32") / ramp_length





from __future__ import annotations
from typing import Optional, List, Tuple, Union, Any, Dict
import tensorflow as tf
from tensorflow.python.util import nest
from returnn.tf.layers.basic import LayerBase
from . import _types
from . import module as _module
from . import naming as _naming
from . import tensor as _tensor
from . import namespace as _namespace


class CallEntry:
  """
  Can be a module() call, or regular func.
  Note that a module can be called multiple times.
  """
  module: _module.ModuleEntry
  orig_inputs_args: Optional[Tuple[Union[_types.Tensor, Any], ...]] = None
  orig_inputs_kwargs: Optional[Dict[str, Union[_types.Tensor, Any]]] = None
  orig_inputs_flat: Optional[List[Union[_types.Tensor, Any]]] = None
  orig_outputs: Optional[Union[_types.Tensor, Tuple[_types.Tensor, ...], Any]] = None
  orig_outputs_flat: Optional[List[Union[_types.Tensor, Any]]] = None
  inputs_args: Optional[Tuple[Union[_tensor.TensorEntry, Any], ...]] = None
  inputs_kwargs: Optional[Dict[str, Union[_tensor.TensorEntry, Any]]] = None
  inputs_flat: Optional[List[Union[_tensor.TensorEntry, int, float, Any]]] = None
  outputs: Optional[Union[_tensor.TensorEntry, Any]] = None
  outputs_flat: Optional[List[Union[_tensor.TensorEntry, Any]]] = None
  parent_call: Optional[CallEntry] = None  # parent in the call stack
  child_calls: List[CallEntry]
  level: Optional[int] = None
  namespace: Optional[_namespace.RegisteredName] = None
  returnn_layer: Optional[LayerBase] = None
  returnn_layer_dict: Optional[Dict[str, Any]] = None

  def __init__(self, module: _module.ModuleEntry):
    self.module = module
    module.calls.append(self)
    self.child_calls = []

  def __repr__(self):
    return (
      f"<{self.__class__.__name__}"
      f" {self.namespace.get_absolute_name() if self.namespace else '?'!r}"
      f" {self.module!r}"
      f" (depth {self.level})>")

  def get_root_call(self) -> CallEntry:
    entry = self
    while entry.parent_call:
      entry = entry.parent_call
    return entry

  def get_canonical_name(self) -> str:
    """
    Considering the canonical context where this is being used.
    Not an absolute name but relative.
    """
    return self.module.get_canonical_name(parent_namespace=self.namespace.parent)

  def set_returnn_layer(self, layer: Optional[LayerBase], layer_dict: Optional[Dict[str, Any]]):
    self.returnn_layer = layer
    self.returnn_layer_dict = layer_dict

  def set_outputs(self, outputs: Union[_types.Tensor, Tuple[_types.Tensor], List[_types.Tensor]]):
    assert self.outputs_flat is None and self.outputs is None
    outputs_flat = nest.flatten(outputs)
    naming = _naming.Naming.get_instance()
    if naming.keep_orig_module_io_tensors:
      self.orig_outputs = outputs
      self.orig_outputs_flat = outputs_flat
    if naming.wrap_to_returnn_enabled:  # not all tensors are traced currently otherwise. also not needed
      # we might get None, e.g. from flattened PackedSequence
      entry_outputs = [naming.tensors[x] if x is not None else None for x in outputs_flat]
      self.outputs_flat = entry_outputs
      self.outputs = nest.pack_sequence_as(structure=outputs, flat_sequence=entry_outputs)
      for x in entry_outputs:
        if x is not None:
          if self not in x.output_from_calls:
            x.output_from_calls.append(self)
            x.output_from_modules.append(self.module)
      if entry_outputs:
        if self.namespace not in entry_outputs[0].names:
          entry_outputs[0].names.append(self.namespace)

  def apply_call(self) -> _types.Tensor:
    from pytorch_to_returnn.torch.nn import Module
    from pytorch_to_returnn.torch import Tensor
    from returnn.tf.util.basic import reuse_name_scope
    naming = _naming.Naming.get_instance()
    module = self.module.module
    assert isinstance(module, Module)
    assert self.namespace
    inputs_flat = [x.tensor() if isinstance(x, _tensor.TensorEntry) else x for x in self.inputs_flat]
    inputs_args, inputs_kwargs = nest.pack_sequence_as(
      structure=(self.inputs_args, self.inputs_kwargs), flat_sequence=inputs_flat)

    if module.has_torch_forward():
      for x in inputs_flat:
        if isinstance(x, Tensor):
          self.namespace.register_input(tensor=naming.tensors[x])
      res = module.forward(*inputs_args, **inputs_kwargs)
      res_flat = nest.flatten(res)
      assert len(res_flat) >= 1
      if self.namespace.returnn_ctx.sub_net_layer:
        res_entry = naming.tensors[res_flat[0]]
        assert isinstance(res_entry, _tensor.TensorEntry)
        self.namespace.register_returnn_subnet_output(res_entry)
        # No need to have separate register logic for the other outputs, if there are multiple.
        # The logic in `name_for_tensor` should find the entry from the subnet.
      layer = self.namespace.returnn_ctx.sub_net_layer
      layer_dict = None  # will be constructed later lazily when needed
      new_update_ops = []  # ignore

    else:  # no module.forward, direct RETURNN layer call
      assert module.create_returnn_layer_dict is not Module.create_returnn_layer_dict
      assert self.namespace and self.namespace.parent
      parent_namespace = self.namespace.parent
      parent_namespace.maybe_create_returnn_ctx()
      layer_dict = module.create_returnn_layer_dict(*inputs_args, **inputs_kwargs)
      layer_name = self.namespace.name
      returnn_net = parent_namespace.returnn_ctx.network
      assert layer_name not in returnn_net.layers
      if len(self.module.calls) >= 2:
        if list(self.module.module.parameters()) or list(self.module.module.buffers()):
          call0 = self.module.calls[0]
          prev_call_name, _ = parent_namespace.name_in_ctx([call0.namespace])
          layer_dict["reuse_params"] = prev_call_name
      print(f"*** {returnn_net.name}/{layer_name!r} layer dict: {layer_dict}")

      # Now the main construction of the layer itself.
      # Collect also potential new TF update ops (e.g. running statistics).
      prev_update_ops = tf.compat.v1.get_collection(tf.compat.v1.GraphKeys.UPDATE_OPS)
      with reuse_name_scope(parent_namespace.returnn_ctx.tf_name_scope, absolute=True):
        layer = returnn_net.construct_layer(net_dict={layer_name: layer_dict}, name=layer_name)
      new_update_ops = tf.compat.v1.get_collection(tf.compat.v1.GraphKeys.UPDATE_OPS)
      assert len(prev_update_ops) <= len(new_update_ops)
      assert all(a is b for (a, b) in zip(prev_update_ops, new_update_ops))
      new_update_ops = new_update_ops[len(prev_update_ops):]

      module.check_returnn_layer(layer)
      res = module.make_output_tensor_from_returnn(inputs_flat=inputs_flat, layer=layer)
      res_entry = naming.tensors[res]
      assert isinstance(res_entry, _tensor.TensorEntry)
      res_entry.returnn_data = layer.output
      self.namespace.assign_tensor(res_entry)
      res_entry.output_from_calls.append(self)  # do now, in case it gets lost after make_structured_returnn_output

      res = module.make_structured_returnn_output(res, *inputs_args, **inputs_kwargs)

    self.set_returnn_layer(layer=layer, layer_dict=layer_dict)
    self.set_outputs(res)

    if layer:  # might not exist in the root namespace
      layer_abs_repr_name = f"{layer.network.name}/{layer.name!r}"
      print(
        f"*** {layer_abs_repr_name} {layer.__class__.__name__} output: "
        f"[{','.join(layer.output.get_batch_axes_short_description())}]")

      if naming.import_params_from_torch_namespace and layer:
        if not layer.get_absolute_name().startswith("."):  # temp layer
          if module.is_original_torch_module and not module.has_torch_forward():
            if list(module.parameters(recurse=False)):
              # Need some way to get a unique name, to get the corresponding module from previous Torch run.
              # We can use get_module_abs_name, to get the attrib chain,
              # or get_module_abs_call_name, to get the call chain,
              # or get_module_abs_id_name as a combination.
              # See the corresponding documentation.
              mod_abs_name = naming.get_module_abs_id_name(module)
              torch_mod = naming.import_params_from_torch_namespace.get_module_by_abs_id_name(mod_abs_name)
              print(
                f"*** {layer_abs_repr_name} {layer.__class__.__name__} "
                f"importing params {[name for name, _ in module.named_parameters(recurse=False)]} ...")
              if layer.reuse_params is None:
                module.import_params_torch_to_returnn(layer=layer, torch_module=torch_mod)

            print(
              f"*** {layer_abs_repr_name} {layer.__class__.__name__} "
              f"check RETURNN inputs/outputs given Torch inputs/outputs ...")
            module.check_call_returnn_outputs_to_prev_torch(self, update_ops=new_update_ops)

    return res

  def __enter__(self):
    # Assume via push_func_call
    assert _naming.Naming.get_instance().module_call_stack[-1] is self
    return self

  def __exit__(self, exc_type, exc_val, exc_tb):
    _naming.Naming.get_instance().pop_module_call(self)



import numpy
from ..modules.rnn import PackedSequence
from ..modules.shape import FlattenBatch
from ...tensor import Tensor
from ..._C import from_numpy
from ....naming import Naming


def pack_padded_sequence(input: Tensor, lengths, batch_first=False, enforce_sorted=True) -> PackedSequence:
  batch_sizes = []
  for frame in range(lengths[0]):
    batch_sizes.append(sum([x > frame for x in lengths]))
  return pack_padded_sequence_with_batch_sizes(input, from_numpy(numpy.array(batch_sizes)), batch_first=batch_first)


def pack_padded_sequence_with_batch_sizes(input: Tensor, batch_sizes: Tensor, batch_first=False) -> PackedSequence:
  # `batch_first` refers to whether the input is a batch major tensor. In the `PackedSequence`, it is always flattened
  # as time major, so we just ignore `batch_first` because RETURNN does not need this information about the input for
  # packing.
  assert batch_first == input.shape[0].is_batch_dim
  return PackedSequence(FlattenBatch(batch_major=False)(input), batch_sizes)


def pad_packed_sequence(sequence: PackedSequence, batch_first=False, padding_value=0.0, total_length=None):
  assert padding_value == 0.0, "not implemented yet"
  assert total_length is None, "not implemented yet"

  return sequence.get_padded_tensor(batch_first=batch_first), sequence.batch_sizes


def pack_sequence(sequences, enforce_sorted=True):
  if isinstance(sequences, Tensor):
    # Note that this is not the intended way to use :func:`pack_sequence`, `sequences` is actually intended to be a list
    # see https://pytorch.org/docs/stable/generated/torch.nn.utils.rnn.pack_sequence.html
    #
    # The function also works with a batch-time-major tensor as input, however, then we cannot have different sequence
    # lengths on the PyTorch side and all sequences are assumed to have the same length on the RETURNN side as well.
    naming = Naming.get_instance()
    tensor_entry = naming.tensors[sequences]
    torch_axis_from_returnn_axis = {j: i for i, j in tensor_entry.returnn_axis_from_torch_axis.items()}
    batch_dim = sequences.shape[torch_axis_from_returnn_axis[tensor_entry.returnn_data.batch_dim_axis]]
    time_dim = sequences.shape[torch_axis_from_returnn_axis[tensor_entry.returnn_data.time_dim_axis]]
    lengths = [time_dim] * batch_dim
    # batch_first is always True because we assume batch-time-major tensor, see above
    return pack_padded_sequence(sequences, lengths, enforce_sorted=enforce_sorted, batch_first=True)
  else:
    raise NotImplementedError


def pad_sequence():
  raise NotImplementedError




from __future__ import annotations
from typing import Optional, List, Tuple
from . import _types
from . import call as _call
from . import namespace as _namespace
from . import naming as _naming


class ModuleEntry:
  module: _types.Module
  level: Optional[int] = None
  calls: List[_call.CallEntry]
  names: List[_namespace.RegisteredName]
  canonical_name: Optional[str] = None
  parent_owning_modules: List[Tuple[ModuleEntry, str]]
  parent_context_modules: List[ModuleEntry]

  def __init__(self, module: _types.Module):
    self.module = module
    self.calls = []
    self.names = []
    self.parent_owning_modules = []
    self.parent_context_modules = []

  def __repr__(self):
    module_repr = repr(self.module)
    # torch.nn.Module.__repr__ can be too verbose when there are children...
    # Strip that away.
    if "\n" in module_repr:
      lines = module_repr.splitlines()
      assert len(lines) >= 2
      module_repr = f"{lines[0].strip()}...{lines[-1].strip()}"
    return f"<ModuleEntry {module_repr}>"

  def get_parent_calling_modules(self) -> List[ModuleEntry]:
    res = []
    for call in self.calls:
      assert call.module is self
      while call.parent_call:
        call = call.parent_call
        if call.module:
          res.append(call.module)
          break
    return res

  def get_root_owning_module(self) -> ModuleEntry:
    mod = self
    while mod.parent_owning_modules:
      mod = mod.parent_owning_modules[0][0]
    return mod

  def get_canonical_name(self, parent_namespace: Optional[_namespace.RegisteredName] = None, *, _visited=None) -> str:
    if self.canonical_name:
      return self.canonical_name
    if _visited is None:
      _visited = set()
    _visited.add(self)
    naming = _naming.Naming.get_instance()
    if parent_namespace is None:
      parent_namespace = naming.root_namespace
    if self.parent_owning_modules:
      mod, name = self.parent_owning_modules[0]
      if parent_namespace and mod in parent_namespace.modules:
        prefix = ""
      elif not mod.module.has_torch_forward() and mod not in _visited:
        prefix = mod.get_canonical_name(_visited=_visited)
        if prefix:
          prefix += "_"
      else:
        prefix = ""
      if not prefix and name[:1].isnumeric():
        return f"layer{name}"
      return prefix + name
    if parent_namespace and self in parent_namespace.modules:
      return self.module.get_returnn_name()
    if set(self.parent_context_modules).intersection(_visited):
      return self.module.get_returnn_name()
    if parent_namespace and parent_namespace is not naming.root_namespace:
      for mod in self.parent_context_modules:
        if mod in parent_namespace.modules:
          return self.module.get_returnn_name()
    prefix = ""
    for mod in reversed(self.parent_context_modules):
      prefix = mod.get_canonical_name(_visited=_visited, parent_namespace=parent_namespace)
      if prefix:
        prefix += "_"
        break
    return prefix + self.module.get_returnn_name()

  def __enter__(self):
    return self

  def __exit__(self, exc_type, exc_val, exc_tb):
    _naming.Naming.get_instance().pop_module_context(self.module)





import _setup_test_env  # noqa
import sys
import unittest
from typing import Dict, Any
import numpy
import contextlib
import tensorflow as tf
from pytorch_to_returnn.naming import Naming
from pytorch_to_returnn import torch
from pytorch_to_returnn.torch import Tensor
from pytorch_to_returnn.pprint import pformat


def run_returnn(module: torch.nn.Module, returnn_data_dict: Dict[str, Any], train_flag=False):
  with make_scope() as session:
    from returnn.config import Config
    from returnn.tf.network import TFNetwork
    config = Config({
      "extern_data": {"data": returnn_data_dict},
      "debug_print_layer_output_template": True,
    })
    net_dict = module.as_returnn_net_dict(returnn_data_dict)
    print()
    print("Net dict:")
    print(pformat(net_dict))
    network = TFNetwork(config=config, name="root", train_flag=train_flag)
    network.construct_from_dict(net_dict)
    out = network.get_default_output_layer().output
    session.run(tf.compat.v1.global_variables_initializer())
    session.run((out.placeholder, out.size_placeholder.as_dict()), feed_dict=make_feed_dict(network.extern_data))


@contextlib.contextmanager
def make_scope():
  """
  :rtype: tf.compat.v1.Session
  """
  with tf.Graph().as_default() as graph:
    with tf.compat.v1.Session(graph=graph) as session:
      yield session


def make_feed_dict(data_list, same_time=False, n_batch=3, n_time=7):
  """
  :param list[returnn.tf.util.data.Data]|ExternData data_list:
  :param bool same_time:
  :param int n_batch:
  :param int n_time:
  :rtype: dict[tf.Tensor,numpy.ndarray]
  """
  from returnn.tf.network import ExternData
  if isinstance(data_list, ExternData):
    data_list = [value for (key, value) in sorted(data_list.data.items())]
  assert n_time > 0 and n_batch > 0
  rnd = numpy.random.RandomState(42)
  existing_sizes = {}  # type: Dict[tf.Tensor,int]
  d = {}
  for data in data_list:
    shape = list(data.batch_shape)
    if data.batch_dim_axis is not None:
      shape[data.batch_dim_axis] = n_batch
    for axis, dim in enumerate(shape):
      if dim is None:
        axis_wo_b = data.get_batch_axis_excluding_batch(axis)
        assert axis_wo_b in data.size_placeholder
        dyn_size = data.size_placeholder[axis_wo_b]
        if dyn_size in existing_sizes:
          shape[axis] = existing_sizes[dyn_size]
          continue
        existing_sizes[dyn_size] = n_time
        shape[axis] = n_time
        dyn_size_v = numpy.array([n_time, max(n_time - 2, 1), max(n_time - 3, 1)])
        if dyn_size_v.shape[0] > n_batch:
          dyn_size_v = dyn_size_v[:n_batch]
        elif dyn_size_v.shape[0] < n_batch:
          dyn_size_v = numpy.concatenate(
            [dyn_size_v, rnd.randint(1, n_time + 1, size=(n_batch - dyn_size_v.shape[0],))], axis=0)
        d[dyn_size] = dyn_size_v
        if not same_time:
          n_time += 1
    print("%r %r: shape %r" % (data, data.placeholder, shape))
    if data.sparse:
      d[data.placeholder] = rnd.randint(0, data.dim or 13, size=shape, dtype=data.dtype)
    else:
      d[data.placeholder] = rnd.normal(size=shape).astype(data.dtype)
  return d


def test_Squeeze():
  class Mod(torch.nn.Module):
    def forward(self, x: Tensor) -> Tensor:
      print("x:", x)
      naming = Naming.get_instance()
      x_meta = naming.get_tensor(x)
      assert x_meta.returnn_axis_from_torch_axis == {0: 0, 1: 1, 2: 2}  # id, no reordering
      squeeze_mod = torch.nn.Squeeze(2)
      y = squeeze_mod(x)
      print("y:", y)
      assert y.shape == x.shape[:2]
      y_meta = naming.get_tensor(y)
      assert y_meta.returnn_axis_from_torch_axis == {0: 0, 1: 1}  # id, no reordering
      return y

  mod = Mod()
  run_returnn(module=mod, returnn_data_dict={"shape": (None, 1)})


if __name__ == "__main__":
  if len(sys.argv) <= 1:
    for k, v in sorted(globals().items()):
      if k.startswith("test_"):
        print("-" * 40)
        print("Executing: %s" % k)
        try:
          v()
        except unittest.SkipTest as exc:
          print("SkipTest:", exc)
        print("-" * 40)
    print("Finished all tests.")
  else:
    assert len(sys.argv) >= 2
    for arg in sys.argv[1:]:
      print("Executing: %s" % arg)
      if arg in globals():
        globals()[arg]()  # assume function and execute
      else:
        eval(arg)  # assume Python code and execute




from .call import CallEntry
from .module import ModuleEntry
from .namespace import RegisteredName
from .naming import Naming
from .returnn_ctx import ReturnnContext
from .tensor import TensorEntry
import os as _os

__doc__ = open(_os.path.dirname(__file__) + "/README.rst").read()



#!/usr/bin/env python3


from typing import Optional, TextIO
import types
import torch
import torch.nn.functional
import pytorch_to_returnn.torch as torch_returnn
import pytorch_to_returnn.torch.nn.functional as torch_returnn_nn_functional


def type_name(obj):
  if isinstance(obj, (types.FunctionType, types.BuiltinFunctionType)):
    return "function"
  if isinstance(obj, type):
    if issubclass(obj, torch.nn.Module):
      return "torch.nn.Module"
    return "class"
  return type(obj).__name__


_IgnoreNameList = {
  # Stuff from typing.
  "TYPE_CHECKING", "Tuple", "Optional", "Union", "Any", "Sequence", "List",
  # Other imports:
  "torch",
  "os",
  "warnings",
}

_IgnoreFullnameList = {
  "torch.os",
  "torch.sys",
  "torch.torch",
  "torch.nn.functional.math",
}


def compare(prefix: str, orig_mod, our_mod):
  # Keep the order as-is. Should be like defined in the source?
  for name, value in vars(orig_mod).items():
    if name.startswith("_") or name in _IgnoreNameList:
      continue
    full_name = f"{prefix}.{name}"
    if full_name in _IgnoreFullnameList:
      continue

    report_available(full_name, value, available=hasattr(our_mod, name))


markdown_output_file = None  # type: Optional[TextIO]


def report_available(name: str, obj, available: bool):
  # Potential available True check marks: ✓✅✓✓☑
  # Potential available False check marks: ✗❎✗✘❌
  avail_s = "[x]" if available else "[ ]"
  print(avail_s, name, f"({type_name(obj)})")
  markdown_output_file.write(
    f"- {avail_s} `{name}` ({type_name(obj)})\n")


def main():
  global markdown_output_file
  markdown_output_file = open("Supported.md", "w")
  markdown_output_file.write("This file is auto-generated and should reflect the current state.\n\n")
  markdown_output_file.write("This is a list of supported PyTorch modules and functions.\n")
  markdown_output_file.write("See also [what is unsupported (on a higher level)](Unsupported.md).\n")
  markdown_output_file.write("\n")

  compare("torch.nn", torch.nn, torch_returnn.nn)
  compare("torch.nn.functional", torch.nn.functional, torch_returnn_nn_functional)
  compare("torch", torch, torch_returnn)

  markdown_output_file.close()


if __name__ == '__main__':
  main()




"""
This wraps both torch.nn.functional and torch.functional, and other __torch_function__ on tensors.

Note that this all maps to modules, which will be temporarily created.
In RETURNN, every operation is a layer.
"""

import numpy
import warnings
from pytorch_to_returnn import torch
from typing import Optional, Union, List, Tuple, Dict, TypeVar, Sequence
from . import modules
from ..tensor import Tensor
from .._C import Size, dtype as _dtype
from ...naming import Naming


_number = Union[int, float, numpy.ndarray, numpy.number]
_size = Union[Size, List[int], Tuple[int, ...]]
_shape_t = Union[int, List[int], Size]
_T = TypeVar("_T")
_default_float_type = "float32"
_builtin_sum = sum


def zeros(*size, out=None, dtype=None, layout=None, device=None, requires_grad=False):
  return Tensor(*size, dtype=dtype)


def ones(*size, out=None, dtype=None, layout=None, device=None, requires_grad=False):
  if size and isinstance(size[0], (tuple, list)):
    assert len(size) == 1
    size = tuple(size[0])
  if not dtype:
    dtype = _default_float_type
  return tensor(numpy.ones(size, dtype=dtype))


def arange(*args, out: Optional[Tensor]=None, dtype: Optional[_dtype]=None, device=None,
           requires_grad: bool=False) -> Tensor:
  assert 1 <= len(args) <= 3
  if len(args) == 3:
    start, end, step = args
  elif len(args) == 2:
    start, end = args
    step = 1
  elif len(args) == 1:
    end, = args
    start = 0
    step = 1
  mod = modules.Range()
  return mod(end, start, step, dtype, False)


def tensor(data, *, dtype=None, device=None, requires_grad=False, pin_memory=False):
  from .._C import from_numpy
  x = from_numpy(data)
  if dtype:
    x = x.type(dtype)
  return x


def cast(input: Union[_T, Tensor, _number], dtype: Union[str, _dtype]) -> Union[_T, Tensor]:
  dtype = _dtype(dtype)
  if dtype == get_dtype(input):
    return input
  return modules.Cast(dtype=dtype)(input)


def cat(tensors, dim=0):
  from .modules.operator import Cat
  return Cat(dim).as_returnn_torch_functional()(*tensors)


def get_dtype(tensor: Union[Tensor, _number]) -> _dtype:
  if isinstance(tensor, Tensor):
    return tensor.dtype
  if isinstance(tensor, int):
    return _dtype("int32")
  if isinstance(tensor, float):
    return _dtype(_default_float_type)
  if isinstance(tensor, (numpy.number, numpy.ndarray)):
    return _dtype(str(tensor.dtype))
  raise TypeError(f"unexpected type {type(tensor)}")


def result_type(tensor1: Union[Tensor, _number], tensor2: Union[Tensor, _number]) -> _dtype:
  # https://pytorch.org/docs/stable/generated/torch.result_type.html
  return promote_types(get_dtype(tensor1), get_dtype(tensor2))


def promote_types(type1: Union[str, _dtype], type2: Union[str, _dtype]) -> _dtype:
  # https://pytorch.org/docs/stable/generated/torch.promote_types.html
  type1 = _dtype(type1)
  type2 = _dtype(type2)
  if type1.category_int != type2.category_int:
    if type1.category_int < type2.category_int:
      type1, type2 = type2, type1
    assert type1.category_int > type2.category_int
    return type1
  assert type1.category_int == type2.category_int
  if type1.bit_size == type2.bit_size:
    assert type1 == type2
    return type1
  if type1.bit_size < type2.bit_size:
    type1, type2 = type2, type1
  assert type1.bit_size > type2.bit_size
  return type1


def as_tensor(data: Union[Tensor, _number],
              dtype: Optional[Union[str, _dtype]] = None,
              device=None) -> Tensor:
  if not isinstance(data, Tensor):
    from .._C import from_numpy
    data = from_numpy(data)
  assert isinstance(data, Tensor)
  if dtype:
    data = cast(data, dtype)
  return data


def is_tensor(obj) -> bool:
  return isinstance(obj, Tensor)


def sum(input: Tensor, dim: Optional[int] = None, dtype: Optional[Union[str, _dtype]] = None) -> Tensor:
  assert dim is not None, "not implemented yet"
  return modules.Reduce(mode="sum", axes=dim)(input)


def add(x: Tensor, y: Tensor) -> Tensor:
  dtype = result_type(x, y)
  return modules.BinaryOperator(kind="add")(cast(x, dtype), cast(y, dtype))


def sub(x: Tensor, y: Tensor) -> Tensor:
  dtype = result_type(x, y)
  return modules.BinaryOperator(kind="sub")(cast(x, dtype), cast(y, dtype))


def mul(x: Tensor, y: Tensor) -> Tensor:
  dtype = result_type(x, y)
  return modules.BinaryOperator(kind="mul")(cast(x, dtype), cast(y, dtype))


def matmul(input: Tensor, other: Tensor, *, out: Optional[Tensor] = None) -> Tensor:
  assert out is None, "not implemented otherwise"
  mod = modules.Matmul()
  return mod(input, other)


def bmm(input: Tensor, mat2: Tensor, *, deterministic: bool = False, out: Optional[Tensor] = None) -> Tensor:
  assert out is None, "not implemented otherwise"
  assert input.shape[0] == mat2.shape[0], "First axis must match"
  assert input.shape[2] == mat2.shape[1], "Axes for matrix-matrix product must match"
  return matmul(input, mat2)


def truediv(x: Tensor, y: Tensor) -> Tensor:
  dtype = result_type(x, y)
  return modules.BinaryOperator(kind="truediv")(cast(x, dtype), cast(y, dtype))


def greater_equal(x: Tensor, y: Tensor) -> Tensor:
  dtype = result_type(x, y)
  return modules.ComparisonOperator(kind="greater_equal")(cast(x, dtype), cast(y, dtype))


def flatten(input: Tensor, start_dim=0, end_dim=-1) -> Tensor:
  return modules.Flatten(start_dim=start_dim, end_dim=end_dim).as_returnn_torch_functional()(input)


def unsqueeze(input: Tensor, dim: int) -> Tensor:
  return input.unsqueeze(dim)


def reshape(input: Tensor, shape: Tuple[int, ...]) -> Tensor:
  if any(dim == -1 for dim in shape):
    num = input.numel()
    for dim in shape:
      if dim == -1:
        continue
      assert dim > 0 and num % dim == 0
      num //= dim
    shape = [dim if dim >= 0 else num for dim in shape]

  # Use Flatten, Unflatten, Squeeze.
  # (Other reshapes are disallowed.)
  axis1, axis2 = 0, 0
  while axis1 < len(input.shape) and axis2 < len(shape):
    if input.shape[axis1] == shape[axis2]:
      axis1 += 1
      axis2 += 1
      continue
    elif input.shape[axis1] < shape[axis2]:
      if input.shape[axis1] == 1:
        input = modules.Squeeze(dim=axis1).as_returnn_torch_functional()(input)
        continue
      n = 1
      a = axis1
      while a < len(input.shape) and n < shape[axis2]:
        assert shape[axis2] % n == 0 and n < shape[axis2]
        n *= input.shape[a]
        a += 1
      if n == shape[axis2]:
        input = modules.Flatten(start_dim=axis1, end_dim=a - 1).as_returnn_torch_functional()(input)
        assert input.shape[axis1] == shape[axis2]
        continue
      elif shape[axis2] % input.shape[axis1] == 0:
        split_factor = shape[axis2] // input.shape[axis1]
        assert input.shape[axis1 + 1] == shape[axis2 + 1] * split_factor
        # something like (..., a, b*c,...) -> (..., a*b, c,...)
        unflattened_size = (split_factor, shape[axis2 + 1])
        input = modules.Unflatten(dim=axis1 + 1, unflattened_size=unflattened_size).as_returnn_torch_functional()(input)
        input = modules.Flatten(start_dim=axis1, end_dim=axis1 + 1).as_returnn_torch_functional()(input)
        assert input.shape[axis1] == shape[axis2]
        axis1 += 1
        axis2 += 1
        assert input.shape[axis1] == shape[axis2]
        continue
    elif input.shape[axis1] > shape[axis2]:
      n = 1
      a = axis2
      while a < len(shape) and n < input.shape[axis1]:
        assert input.shape[axis1] % n == 0 and n < input.shape[axis1]
        n *= shape[a]
        a += 1
      if n == input.shape[axis1]:
        unflattened_size = tuple(shape[axis2:a])
        input = modules.Unflatten(dim=axis1, unflattened_size=unflattened_size).as_returnn_torch_functional()(input)
        assert input.shape[axis1] == shape[axis2]
        continue
      elif input.shape[axis1] % shape[axis2] == 0:
        split_factor = input.shape[axis1] // shape[axis2]
        assert input.shape[axis1 + 1] * split_factor == shape[axis2 + 1]
        # something like (..., a*b, c,...) -> (..., a, b*c,...)
        unflattened_size = (shape[axis2], split_factor)
        input = modules.Unflatten(dim=axis1, unflattened_size=unflattened_size).as_returnn_torch_functional()(input)
        assert input.shape[axis1] == shape[axis2]
        axis1 += 1
        axis2 += 1
        input = modules.Flatten(start_dim=axis1, end_dim=axis1 + 1).as_returnn_torch_functional()(input)
        assert input.shape[axis1] == shape[axis2]
        continue
    else:
      assert False  # cannot happen
  assert axis1 == axis2
  if len(input.shape) < len(shape):
    assert all(shape[i] == 1 for i in range(len(input.shape), len(shape)))
    input = modules.Unflatten(
      dim=-1, unflattened_size=shape[len(input.shape) - 1:]).as_returnn_torch_functional()(input)
  elif len(input.shape) > len(shape):
    while len(input.shape) > len(shape):
      input = modules.Squeeze(dim=len(shape)).as_returnn_torch_functional()(input)
  assert len(input.shape) == len(shape) and input.shape == tuple(shape)
  return input


def movedim(input: Tensor, source: Union[int, Tuple[int, ...]], destination: Union[int, Tuple[int, ...]]):
  if isinstance(source, int):
    source = (source,)
  if isinstance(destination, int):
    destination = (destination,)
  assert isinstance(source, (tuple, list)) and isinstance(destination, (tuple, list))
  assert len(source) == len(destination)
  perm = {i: j for i, j in zip(destination, source)}
  # All remaining axes stay in order.
  return tensorflow_transpose(input, perm=perm)


def transpose(input: Tensor, dim0: int, dim1: int):
  return tensorflow_transpose(input, perm={dim0: dim1, dim1: dim0})


def permute(input: Tensor, dims: Tuple[int]):
  return tensorflow_transpose(input, perm=dims)


def t(input: Tensor):
  if len(input.shape) < 2:
    return input
  elif len(input.shape) == 2:
    return transpose(input, 0, 1)
  else:
    # https://pytorch.org/docs/stable/generated/torch.t.html#torch.t
    raise ValueError("t() expects input to be <= 2-D tensor")


def tensorflow_transpose(input: Tensor, perm: Optional[Union[Dict[int, int], Tuple[int, ...], List[int]]]):
  """
  Note: This function is added by us, not available in original PyTorch.

  Note: The resulting Torch tensor is transposed as expected.
  However, on the RETURNN side, we actually should never need to transpose,
  as we have dimension tags, and all layers should refer to axes by dim tags.
  So on RETURNN side, this is a no-op.
  """
  return modules.Transpose(perm=perm)(input)


def expand(input: Tensor, *sizes: _size):
  if sizes and isinstance(sizes[0], (tuple, list)):
    assert len(sizes) == 1
    shape = tuple(sizes[0])
  else:
    shape = sizes
  assert isinstance(shape, tuple)

  # reshape to add new dims
  if len(shape) != len(input.shape):
    assert len(shape) >= len(input.shape), "Cannot reduce dimensions"
    assert -1 not in [d for d in shape[:-len(input.shape)]], "Size -1 is not allowed for expanded dims"
    exp_shape = (1,) * (len(shape) - len(input.shape)) + input.shape[-len(input.shape):]
    input = reshape(input, shape=exp_shape)

  # tile to expand to larger size
  multiples = {}
  for axis, multiple in enumerate(shape):
    if input.shape[axis] > 1:
      assert multiple == -1 or multiple == input.shape[axis], \
        f"Expanded size ({multiple}) must match existing size ({input.shape[axis]}) at non-singleton dimension ({axis})"
      multiple = 1
    if multiple == -1:
      multiple = 1  # -1 as the size for a dim means not changing the size of that dim
    multiples[axis] = multiple
  return modules.Tile(multiples=multiples).as_returnn_torch_functional()(input)


def chunk(input: Tensor, chunks: int, dim: int = 0) -> List[Tensor]:
  chunk_size = input.shape[dim] // chunks
  if input.shape[dim] % chunks != 0:
    chunk_size += 1
    size_splits = [chunk_size] * chunks
    size_splits[-1] -= chunk_size * chunks - input.shape[dim]
  else:
    size_splits = [chunk_size] * chunks
  assert _builtin_sum(size_splits) == input.shape[dim] and len(size_splits) == chunks
  return modules.Split(dim=dim, size_splits=size_splits).as_returnn_torch_functional()(input)


def pad(input: Tensor, pad, mode='constant', value=0) -> Tensor:
  return modules.GenericPadNd(padding=pad, mode=mode, value=value).as_returnn_torch_functional()(input)


def max(*inputs: Tensor) -> Tensor:
  return modules.Max()(*inputs)


def linear(input: Tensor, weight: Tensor, bias: Optional[Tensor] = None):
  output = input.matmul(weight.t())
  if bias is not None:
    output += bias
  return output


def conv1d(
    input: Tensor, weight: Tensor, bias: Optional[Tensor] = None,
    stride: Union[int, _size] = 1, padding: Union[int, _size] = 0,
    dilation: Union[int, _size] = 1, groups: int = 1) -> Tensor:
  mod = modules.FunctionalConv1d(stride=stride, padding=padding, dilation=dilation, groups=groups)
  return mod(input, weight, bias)


def conv2d(
    input: Tensor, weight: Tensor, bias: Optional[Tensor] = None,
    stride: Union[int, _size] = 1, padding: Union[int, _size] = 0,
    dilation: Union[int, _size] = 1, groups: int = 1) -> Tensor:
  mod = modules.FunctionalConv2d(stride=stride, padding=padding, dilation=dilation, groups=groups)
  return mod(input, weight, bias)


def conv_transpose1d(
    input: Tensor, weight: Tensor, bias: Optional[Tensor] = None,
    stride: Union[int, _size] = 1, padding: Union[int, _size] = 0,
    output_padding: Union[int, _size] = 0,
    groups: int = 1,
    dilation: Union[int, _size] = 1) -> Tensor:
  mod = modules.FunctionalConvTransposed1d(
    stride=stride, padding=padding, output_padding=output_padding, dilation=dilation, groups=groups)
  return mod(input, weight, bias)


def max_pool2d(input: Tensor, kernel_size, stride=None, padding=0, dilation=1,
               ceil_mode=False, return_indices=False):
  mod = modules.MaxPool2d(
    kernel_size=kernel_size, stride=stride, padding=padding, dilation=dilation,
    ceil_mode=ceil_mode, return_indices=return_indices)
  mod.as_returnn_torch_functional()
  return mod(input)


def relu(input: Tensor) -> Tensor:
  return modules.ReLU().as_returnn_torch_functional()(input)


def gelu(input: Tensor) -> Tensor:
  return modules.GELU().as_returnn_torch_functional()(input)


def leaky_relu(input: Tensor, negative_slope: float = 0.01, inplace: bool = False) -> Tensor:
  return modules.LeakyReLU(negative_slope=negative_slope, inplace=inplace).as_returnn_torch_functional()(input)


def sqrt(input: Tensor) -> Tensor:
  return modules.Sqrt().as_returnn_torch_functional()(input)


def tanh(input: Tensor) -> Tensor:
  return modules.Tanh().as_returnn_torch_functional()(input)


def softmax(input: Tensor, dim: Optional[int] = None, dtype=None):
  return modules.Softmax(dim=dim).as_returnn_torch_functional()(input)


def abs(input: Tensor) -> Tensor:
  return modules.Abs().as_returnn_torch_functional()(input)


def log_softmax(input: Tensor, dim: Optional[int] = None, dtype=None):
  return modules.LogSoftmax(dim=dim).as_returnn_torch_functional()(input)


def log(input: Tensor):
  return modules.Log().as_returnn_torch_functional()(input)


def log1p(input, *, out=None):
  return log(input + 1)


def sigmoid(input: Tensor):
  return modules.Sigmoid().as_returnn_torch_functional()(input)


def logsigmoid(input: Tensor):
  return modules.LogSigmoid().as_returnn_torch_functional()(input)


def pow(input: Tensor, exponent: float):
  return modules.Power(exponent=exponent).as_returnn_torch_functional()(input)


def normalize(input: Tensor, p=2, dim=1, eps=1e-12) -> Tensor:
  norm_ = modules.Norm(p=p, axes=[dim], keepdims=True)(input)
  norm_f = modules.Reciprocal(eps=eps)(norm_)
  return input * norm_f


def norm(input: Tensor,
         p: Optional[Union[str, float, int]] = "fro",
         dim: Optional[Union[int, List[int]]] = None,
         keepdim: bool = False) -> Tensor:
  return modules.Norm(p=p, axes=[dim], keepdims=keepdim)(input)


def norm_except_dim(v: Tensor, pow: int = 2, dim: int = 0) -> Tensor:
  return modules.Norm(p=pow, axes=[i for i in range(v.dim()) if i != dim], keepdims=True)(v)


def layer_norm(input: Tensor, normalized_shape: _shape_t, weight: Optional[Tensor] = None,
               bias: Optional[Tensor] = None, eps: float = 1e-5) -> Tensor:
  module = modules.LayerNorm(normalized_shape=normalized_shape, eps=eps, elementwise_affine=False)
  out = module.as_returnn_torch_functional()(input)
  if weight is not None:
    assert (out.shape[-1],) == weight.shape, "data should be of shape (B, *, F) and weights should be of shape (F,)"
    out *= weight
  if bias is not None:
    assert (out.shape[-1],) == bias.shape, "data should be of shape (B, *, F) and bias should be of shape (F,)"
    out += bias
  return out


def group_norm(input: Tensor, num_groups: int, weight: Optional[Tensor] = None, bias: Optional[Tensor] = None,
               eps: float = 1e-5) -> Tensor:
  module = modules.GroupNorm(num_groups=num_groups, num_channels=input.shape[1], eps=eps, affine=False)
  out = module.as_returnn_torch_functional()(input)
  if weight is not None:
    assert (out.shape[1],) == weight.shape, "data should be of shape (B, F, *) and weights should be of shape (F,)"
    out *= weight.view(1, -1, 1)
  if bias is not None:
    assert (out.shape[1],) == bias.shape, "data should be of shape (B, F, *) and bias should be of shape (F,)"
    out += bias.view(1, -1, 1)
  return out


def dropout(input: Tensor, p: float = 0.5, training: bool = True, inplace: bool = False) -> Tensor:
  if p < 0. or p > 1.:
    raise ValueError("dropout probability has to be between 0 and 1, but got {}".format(p))
  if p == 0.0 or not training:
    return input
  if p > 0.0:
    return modules.Dropout(p=p, inplace=inplace).as_returnn_torch_functional()(input)


def multi_head_attention_forward(
  query: Tensor,
  key: Tensor,
  value: Tensor,
  embed_dim_to_check: int,
  num_heads: int,
  in_proj_weight: Tensor,
  in_proj_bias: Optional[Tensor],
  bias_k: Optional[Tensor],
  bias_v: Optional[Tensor],
  add_zero_attn: bool,
  dropout_p: float,
  out_proj_weight: Tensor,
  out_proj_bias: Optional[Tensor],
  training: bool = True,
  key_padding_mask: Optional[Tensor] = None,
  need_weights: bool = True,
  attn_mask: Optional[Tensor] = None,
  use_separate_proj_weight: bool = False,
  q_proj_weight: Optional[Tensor] = None,
  k_proj_weight: Optional[Tensor] = None,
  v_proj_weight: Optional[Tensor] = None,
  static_k: Optional[Tensor] = None,
  static_v: Optional[Tensor] = None,
) -> Tuple[Tensor, Optional[Tensor]]:
  # pytorch-to-returnn: remove possible `return handle_torch_function(...)` from original implementation

  tgt_len, bsz, embed_dim = query.size()
  assert embed_dim == embed_dim_to_check
  # allow MHA to have different sizes for the feature dimension
  assert key.size(0) == value.size(0) and key.size(1) == value.size(1)

  if isinstance(embed_dim, torch.Tensor):
    # embed_dim can be a tensor when JIT tracing
    head_dim = embed_dim.div(num_heads, rounding_mode='trunc')
  else:
    head_dim = embed_dim // num_heads
  assert head_dim * num_heads == embed_dim, "embed_dim must be divisible by num_heads"
  scaling = float(head_dim) ** -0.5

  if not use_separate_proj_weight:
    if (query is key or torch.equal(query, key)) and (key is value or torch.equal(key, value)):
      # self-attention
      q, k, v = linear(query, in_proj_weight, in_proj_bias).chunk(3, dim=-1)

    elif key is value or torch.equal(key, value):
      # encoder-decoder attention
      # This is inline in_proj function with in_proj_weight and in_proj_bias
      _b = in_proj_bias
      _start = 0
      _end = embed_dim
      _w = in_proj_weight[_start:_end, :]
      if _b is not None:
        _b = _b[_start:_end]
      q = linear(query, _w, _b)

      if key is None:
        assert value is None
        k = None
        v = None
      else:

        # This is inline in_proj function with in_proj_weight and in_proj_bias
        _b = in_proj_bias
        _start = embed_dim
        _end = None
        _w = in_proj_weight[_start:, :]
        if _b is not None:
          _b = _b[_start:]
        k, v = linear(key, _w, _b).chunk(2, dim=-1)

    else:
      # This is inline in_proj function with in_proj_weight and in_proj_bias
      _b = in_proj_bias
      _start = 0
      _end = embed_dim
      _w = in_proj_weight[_start:_end, :]
      if _b is not None:
        _b = _b[_start:_end]
      q = linear(query, _w, _b)

      # This is inline in_proj function with in_proj_weight and in_proj_bias
      _b = in_proj_bias
      _start = embed_dim
      _end = embed_dim * 2
      _w = in_proj_weight[_start:_end, :]
      if _b is not None:
        _b = _b[_start:_end]
      k = linear(key, _w, _b)

      # This is inline in_proj function with in_proj_weight and in_proj_bias
      _b = in_proj_bias
      _start = embed_dim * 2
      _end = None
      _w = in_proj_weight[_start:, :]
      if _b is not None:
        _b = _b[_start:]
      v = linear(value, _w, _b)
  else:
    q_proj_weight_non_opt = torch.jit._unwrap_optional(q_proj_weight)
    len1, len2 = q_proj_weight_non_opt.size()
    assert len1 == embed_dim and len2 == query.size(-1)

    k_proj_weight_non_opt = torch.jit._unwrap_optional(k_proj_weight)
    len1, len2 = k_proj_weight_non_opt.size()
    assert len1 == embed_dim and len2 == key.size(-1)

    v_proj_weight_non_opt = torch.jit._unwrap_optional(v_proj_weight)
    len1, len2 = v_proj_weight_non_opt.size()
    assert len1 == embed_dim and len2 == value.size(-1)

    if in_proj_bias is not None:
      q = linear(query, q_proj_weight_non_opt, in_proj_bias[0:embed_dim])
      k = linear(key, k_proj_weight_non_opt, in_proj_bias[embed_dim: (embed_dim * 2)])
      v = linear(value, v_proj_weight_non_opt, in_proj_bias[(embed_dim * 2):])
    else:
      q = linear(query, q_proj_weight_non_opt, in_proj_bias)
      k = linear(key, k_proj_weight_non_opt, in_proj_bias)
      v = linear(value, v_proj_weight_non_opt, in_proj_bias)
  q = q * scaling

  if attn_mask is not None:
    assert (
      attn_mask.dtype == torch.float32
      or attn_mask.dtype == torch.float64
      or attn_mask.dtype == torch.float16
      or attn_mask.dtype == torch.uint8
      or attn_mask.dtype == torch.bool
    ), "Only float, byte, and bool types are supported for attn_mask, not {}".format(attn_mask.dtype)
    if attn_mask.dtype == torch.uint8:
      warnings.warn("Byte tensor for attn_mask in nn.MultiheadAttention is deprecated. Use bool tensor instead.")
      attn_mask = attn_mask.to(torch.bool)

    if attn_mask.dim() == 2:
      attn_mask = attn_mask.unsqueeze(0)
      if list(attn_mask.size()) != [1, query.size(0), key.size(0)]:
        raise RuntimeError("The size of the 2D attn_mask is not correct.")
    elif attn_mask.dim() == 3:
      if list(attn_mask.size()) != [bsz * num_heads, query.size(0), key.size(0)]:
        raise RuntimeError("The size of the 3D attn_mask is not correct.")
    else:
      raise RuntimeError("attn_mask's dimension {} is not supported".format(attn_mask.dim()))
    # attn_mask's dim is 3 now.

  # convert ByteTensor key_padding_mask to bool
  if key_padding_mask is not None and key_padding_mask.dtype == torch.uint8:
    warnings.warn(
      "Byte tensor for key_padding_mask in nn.MultiheadAttention is deprecated. Use bool tensor instead."
    )
    key_padding_mask = key_padding_mask.to(torch.bool)

  if bias_k is not None and bias_v is not None:
    if static_k is None and static_v is None:
      k = torch.cat([k, bias_k.repeat(1, bsz, 1)])
      v = torch.cat([v, bias_v.repeat(1, bsz, 1)])
      if attn_mask is not None:
        attn_mask = pad(attn_mask, (0, 1))
      if key_padding_mask is not None:
        key_padding_mask = pad(key_padding_mask, (0, 1))
    else:
      assert static_k is None, "bias cannot be added to static key."
      assert static_v is None, "bias cannot be added to static value."
  else:
    assert bias_k is None
    assert bias_v is None

  # Unlike the original implementation, we keep batch axis and num_heads separate.
  # Therefore, reshape q, k and v to (bsz, num_heads, tgt_len, head_dim)
  q = q.contiguous().view(tgt_len, bsz, num_heads, head_dim).transpose(0, 1).transpose(1, 2)
  if k is not None:
    k = k.contiguous().view(-1, bsz, num_heads, head_dim).transpose(0, 1).transpose(1, 2)
  if v is not None:
    v = v.contiguous().view(-1, bsz, num_heads, head_dim).transpose(0, 1).transpose(1, 2)

  if static_k is not None:
    assert static_k.size(0) == bsz * num_heads
    assert static_k.size(2) == head_dim
    k = static_k.view(bsz, num_heads, -1, head_dim)  # adapt to separate batch axis

  if static_v is not None:
    assert static_v.size(0) == bsz * num_heads
    assert static_v.size(2) == head_dim
    v = static_v.view(bsz, num_heads, -1, head_dim)  # adapt to separate batch axis

  src_len = k.size(2)  # adapt to separate batch axis

  if key_padding_mask is not None:
    assert key_padding_mask.size(0) == bsz
    assert key_padding_mask.size(1) == src_len

  if add_zero_attn:
    src_len += 1
    # adapt to separate batch axis
    k = torch.cat([k, torch.zeros(k.size()[:2] + (1, k.size(3)), dtype=k.dtype, device=k.device)], dim=2)
    v = torch.cat([v, torch.zeros(v.size()[:2] + (1, v.size(3)), dtype=v.dtype, device=v.device)], dim=2)
    if attn_mask is not None:
      attn_mask = pad(attn_mask, (0, 1))
    if key_padding_mask is not None:
      key_padding_mask = pad(key_padding_mask, (0, 1))

  # adapt to separate batch axis
  attn_output_weights = torch.matmul(q, k.transpose(2, 3))
  assert list(attn_output_weights.size()) == [bsz, num_heads, tgt_len, src_len]

  if attn_mask is not None:
    # adapt to separate batch axis
    attn_mask = attn_mask.view(bsz, num_heads, tgt_len, src_len)
    if attn_mask.dtype == torch.bool:
      attn_output_weights.masked_fill_(attn_mask, float("-inf"))
    else:
      attn_output_weights += attn_mask

  if key_padding_mask is not None:
    # adapt to separate batch axis
    attn_output_weights = attn_output_weights.masked_fill(
      key_padding_mask.unsqueeze(1).unsqueeze(2),
      float("-inf"),
    )

  attn_output_weights = softmax(attn_output_weights, dim=-1)
  attn_output_weights = dropout(attn_output_weights, p=dropout_p, training=training)

  # adapt to separate batch axis
  attn_output = torch.matmul(attn_output_weights, v)
  assert list(attn_output.size()) == [bsz, num_heads, tgt_len, head_dim]
  attn_output = attn_output.transpose(1, 2).transpose(0, 1).contiguous().view(tgt_len, bsz, embed_dim)
  attn_output = linear(attn_output, out_proj_weight, out_proj_bias)

  if need_weights:
    # average attention weights over heads
    attn_output_weights = attn_output_weights.view(bsz, num_heads, tgt_len, src_len)
    return attn_output, attn_output_weights.sum(dim=1) / num_heads
  else:
    return attn_output, None


def mse_loss():
  raise NotImplementedError


def ceil():
  raise NotImplementedError


def clamp():
  raise NotImplementedError


def exp():
  raise NotImplementedError


def log10():
  raise NotImplementedError




import _setup_test_env  # noqa
import sys
import unittest
import typing
import numpy
import tensorflow as tf
from pytorch_to_returnn import torch
from pytorch_to_returnn.converter import verify_torch_and_convert_to_returnn
from pytorch_to_returnn.pprint import pformat


def test_embedding():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    model = torch.nn.Embedding(n_in, n_out)
    return model(inputs)

  rnd = numpy.random.RandomState(42)
  x = rnd.randint(0, n_in, (n_time, n_batch), dtype="int64")
  verify_torch_and_convert_to_returnn(
    model_func,
    inputs=x, inputs_data_kwargs={"shape": (None,), "sparse": True, "dim": n_in, "batch_dim_axis": 1})


def test_linear_multiple_steps():
  n_steps = 3
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    ins = inputs.chunk(n_steps, dim=-1)
    model = torch.nn.Linear(n_in, n_out)
    outs = [model(x) for x in ins]
    out = sum(outs)
    return out

  x = numpy.ones((n_batch, n_time, n_in * n_steps)).astype("float32")
  verify_torch_and_convert_to_returnn(
    model_func, inputs=x, inputs_data_kwargs={
      "shape": (None, n_in * n_steps), "batch_dim_axis": 0, "time_dim_axis": 1, "feature_dim_axis": 2})


def test_load_params_in_returnn():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
      import torch.nn.functional as F
    else:
      torch = wrapped_import("torch")
      F = wrapped_import("torch.nn.functional")

    class DummyModule(torch.nn.Module):
      def __init__(self):
        super(DummyModule, self).__init__()
        self.model = torch.nn.Linear(n_in, n_out)

      def forward(self, x):
        return F.linear(x, self.model.weight)

    mod = DummyModule()
    return mod(inputs)

  x = numpy.ones((n_batch, n_time, n_in)).astype("float32")
  verify_torch_and_convert_to_returnn(
    model_func, inputs=x, inputs_data_kwargs={
      "shape": (None, n_in), "batch_dim_axis": 0, "time_dim_axis": 1, "feature_dim_axis": 2})


def test_load_params_in_returnn_with_initializer():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
      import torch.nn.functional as F
    else:
      torch = wrapped_import("torch")
      F = wrapped_import("torch.nn.functional")

    class DummyModule(torch.nn.Module):
      def __init__(self):
        super(DummyModule, self).__init__()
        self.model = torch.nn.Linear(n_in, n_out)
        torch.nn.init.xavier_uniform_(self.model.weight)

      def forward(self, x):
        return F.linear(x, self.model.weight)

    mod = DummyModule()
    return mod(inputs)

  x = numpy.ones((n_batch, n_time, n_in)).astype("float32")
  verify_torch_and_convert_to_returnn(
    model_func, inputs=x, inputs_data_kwargs={
      "shape": (None, n_in), "batch_dim_axis": 0, "time_dim_axis": 1, "feature_dim_axis": 2})


def test_cat():

  def model_func(wrapped_import, inputs: torch.Tensor):
    if wrapped_import:
      torch = wrapped_import("torch")
    else:
      import torch
    return torch.cat((inputs, inputs), dim=-1)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (3, 3)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_cat_non_feature():
  n_batch, n_time, n_feat = 3, 5, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if wrapped_import:
      torch = wrapped_import("torch")
    else:
      import torch

    x = inputs.expand(1, n_batch, n_feat, n_time)
    y = inputs.expand(3, n_batch, n_feat, n_time)
    return torch.cat([x, y], dim=0)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_feat, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_conv():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    # {'class': 'transposed_conv', 'from': 'layer2', 'activation': None, 'with_bias': True,
    #  'n_out': 192, 'filter_size': (10,), 'strides': (5,), 'remove_padding': (3,), 'output_padding': (1,)}
    model = torch.nn.Conv1d(
      in_channels=n_in,
      out_channels=n_out,
      kernel_size=3,
      stride=2)
    return model(inputs)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_conv2d():
  n_in, n_out = 11, 13
  n_batch, n_time1, n_time2 = 3, 17, 19

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    # {'class': 'transposed_conv', 'from': 'layer2', 'activation': None, 'with_bias': True,
    #  'n_out': 192, 'filter_size': (10,), 'strides': (5,), 'remove_padding': (3,), 'output_padding': (1,)}
    model = torch.nn.Conv2d(
      in_channels=n_in,
      out_channels=n_out,
      kernel_size=(3, 5),
      stride=2)
    return model(inputs)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time1, n_time2)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_conv_transposed():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    # {'class': 'transposed_conv', 'from': 'layer2', 'activation': None, 'with_bias': True,
    #  'n_out': 192, 'filter_size': (10,), 'strides': (5,), 'remove_padding': (3,), 'output_padding': (1,)}
    model = torch.nn.ConvTranspose1d(
      in_channels=n_in,
      out_channels=n_out,
      kernel_size=10,
      stride=5,
      padding=3,
      output_padding=1)
    return model(inputs)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_conv_transposed_2d():
  n_in, n_out = 11, 13
  n_batch, n_time1, n_time2 = 3, 17, 19

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    model = torch.nn.ConvTranspose2d(
      in_channels=n_in,
      out_channels=n_out,
      kernel_size=(10, 3),
      stride=5,
      padding=(2, 3),
      output_padding=1)
    return model(inputs)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time1, n_time2)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_conv_transposed_2d_with_unsqueeze():
  n_in, n_out = 16, 16
  n_batch, n_features, n_time = 4, 16, 20

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    inputs = inputs.unsqueeze(-1)  # (B, F, T, 1)
    model = torch.nn.ConvTranspose2d(
      in_channels=n_in,
      out_channels=n_out,
      kernel_size=(1, 12))
    return model(inputs)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_features, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_functional_linear():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
      import torch.nn.functional as F
    else:
      torch = wrapped_import("torch")
      F = wrapped_import("torch.nn.functional")
    rnd = numpy.random.RandomState(42)
    weight = rnd.normal(0., 1., (n_out, n_in)).astype("float32")
    bias = rnd.normal(0., 1., (n_out,)).astype("float32")
    weight = torch.from_numpy(weight)
    bias = torch.from_numpy(bias)
    return F.linear(inputs.transpose(1, 2), weight=weight, bias=bias)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_matmul_broadcasting():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
      import torch.nn.functional as F
    else:
      torch = wrapped_import("torch")
      F = wrapped_import("torch.nn.functional")
    rnd = numpy.random.RandomState(42)
    weight = rnd.normal(0., 1., (n_in, n_out)).astype("float32")
    weight = torch.from_numpy(weight)
    return torch.matmul(inputs.transpose(1, 2), weight)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_matmul_shared_remaining_axes():
  n_1, n_2 = 2, 4
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    return torch.matmul(inputs, inputs.transpose(2, 3))

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_1, n_time, n_2)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x, inputs_data_kwargs={
      "shape": (n_1, None, n_2), "batch_dim_axis": 0, "time_dim_axis": 2, "feature_dim_axis": 3})


def test_spatial_axes_with_same_tag():
  n_1, n_2 = 2, 4
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
      import torch.nn.functional as F
    else:
      torch = wrapped_import("torch")
      F = wrapped_import("torch.nn.functional")
    x = torch.matmul(inputs, inputs.transpose(2, 3))
    x = F.softmax(x, dim=-1)
    return x

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_1, n_time, n_2)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x, inputs_data_kwargs={
      "shape": (n_1, None, n_2), "batch_dim_axis": 0, "time_dim_axis": 2, "feature_dim_axis": 3})


def test_bmm():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 5

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")

    inputs = inputs.transpose(0, 1)
    x = inputs.new_zeros(inputs.shape[0], inputs.shape[2], n_out) + 1  # (B, F_in, F_out)
    return torch.bmm(inputs, x)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_time, n_batch, n_in)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x, inputs_data_kwargs={
      "shape": (None, n_in), "time_dim_axis": 0, "batch_dim_axis": 1, "feature_dim_axis": 2})


def test_packed_sequence_1():
  """
  Regular packing and unpacking from batched, padded tensor
  """
  n_batch, n_time, n_feat = 3, 5, 7
  seq_lens = [5, 4, 3]

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")

    h = torch.nn.utils.rnn.pack_padded_sequence(inputs, seq_lens)
    output, _ = torch.nn.utils.rnn.pad_packed_sequence(h)
    return output + 1

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_time, n_batch, n_feat)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x, returnn_input_seq_lens={0: seq_lens}, inputs_data_kwargs={
      "shape": (None, n_feat), "time_dim_axis": 0, "batch_dim_axis": 1, "feature_dim_axis": 2})


def test_packed_sequence_2():
  """
  Packing and unpacking from batched, padded tensor, where the packing is done with :func:`pack_sequence` which actually
  requires a list of tensors as input.
  """
  n_batch, n_time, n_feat = 3, 5, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")

    h = torch.nn.utils.rnn.pack_sequence(inputs)
    output, _ = torch.nn.utils.rnn.pad_packed_sequence(h, batch_first=True)
    return output + 1

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_time, n_feat)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x, inputs_data_kwargs={
      "shape": (None, n_feat), "batch_dim_axis": 0, "time_dim_axis": 1, "feature_dim_axis": 2})


def test_packed_sequence_3():
  """
  Pack and return packed .data
  """
  n_batch, n_time, n_feat = 3, 5, 7
  seq_lens = [5, 4, 3]

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")

    h = torch.nn.utils.rnn.pack_padded_sequence(inputs, seq_lens)
    return h.data

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_time, n_batch, n_feat)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x, returnn_input_seq_lens={0: seq_lens}, inputs_data_kwargs={
      "shape": (None, n_feat), "time_dim_axis": 0, "batch_dim_axis": 1, "feature_dim_axis": 2})


def test_packed_sequence_4():
  """
  Initialize :class:`PackedSequence` directly using its init
  """
  n_batch, n_time, n_feat = 3, 5, 7
  seq_lens = [5, 4, 3]

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")

    h = torch.nn.utils.rnn.pack_padded_sequence(inputs, seq_lens)
    h = torch.nn.utils.rnn.PackedSequence(h.data, h.batch_sizes)
    return h.data

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_time, n_batch, n_feat)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x, returnn_input_seq_lens={0: seq_lens}, inputs_data_kwargs={
      "shape": (None, n_feat), "time_dim_axis": 0, "batch_dim_axis": 1, "feature_dim_axis": 2})


def test_packed_sequence_5():
  """
  Pack and return packed .data, like :func:`test_packed_sequence_3` but with batch major input
  """
  n_batch, n_time, n_feat = 3, 5, 7
  seq_lens = [5, 4, 3]

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")

    h = torch.nn.utils.rnn.pack_padded_sequence(inputs, seq_lens, batch_first=True)
    return h.data

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_time, n_feat)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x, returnn_input_seq_lens={0: seq_lens}, inputs_data_kwargs={
      "shape": (None, n_feat), "batch_dim_axis": 0, "time_dim_axis": 1, "feature_dim_axis": 2})


def test_packed_sequence_6():
  """
  Pack with `batch_first=False` and unpack with `batch_first=True`
  """
  n_batch, n_time, n_feat = 3, 5, 7
  seq_lens = [5, 4, 3]

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")

    h = torch.nn.utils.rnn.pack_padded_sequence(inputs, seq_lens)
    output, _ = torch.nn.utils.rnn.pad_packed_sequence(h, batch_first=True)
    return output + 1

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_time, n_batch, n_feat)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x, returnn_input_seq_lens={0: seq_lens}, inputs_data_kwargs={
      "shape": (None, n_feat), "time_dim_axis": 0, "batch_dim_axis": 1, "feature_dim_axis": 2})


def test_lstm_with_packed_sequence_input():
  n_batch, n_time, n_feat = 3, 5, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")

    h = torch.nn.utils.rnn.pack_padded_sequence(inputs, [n_time] * n_batch)
    output, _ = torch.nn.LSTM(n_feat, n_feat)(h)
    return output.data

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_time, n_batch, n_feat)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x, inputs_data_kwargs={
      "shape": (None, n_feat), "time_dim_axis": 0, "batch_dim_axis": 1, "feature_dim_axis": 2})


def test_t():
  n_batch, n_feature, n_time = 3, 5, 17

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
      import torch.nn.functional as F
    else:
      torch = wrapped_import("torch")
      F = wrapped_import("torch.nn.functional")
    rnd = numpy.random.RandomState(42)
    weight = rnd.normal(0., 1., (3, 5)).astype("float32")
    weight = torch.from_numpy(weight)
    weight = weight.t()
    return F.relu(inputs)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_feature, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_reshape_ab_c_to_a_bc():
  n_batch, n_time, n_feature_1, n_feature_2 = 3, 5, 8, 6

  def model_func(wrapped_import, inputs: torch.Tensor):
    # test case (..., a*b, c,...) -> (..., a, b*c,...)
    return inputs.view(n_batch, n_time, n_feature_1 // 2, n_feature_2 * 2)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_time, n_feature_1, n_feature_2)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x, inputs_data_kwargs={
      "shape": (None, n_feature_1, n_feature_2), "batch_dim_axis": 0, "time_dim_axis": 1, "feature_dim_axis": 2})


def test_reshape_a_bc_to_ab_c():
  n_batch, n_time, n_feature_1, n_feature_2 = 3, 5, 8, 6

  def model_func(wrapped_import, inputs: torch.Tensor):
    # test case (..., a, b*c,...) -> (..., a*b, c,...)
    return inputs.view(n_batch, n_time, n_feature_1 * 2, n_feature_2 // 2)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_time, n_feature_1, n_feature_2)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x, inputs_data_kwargs={
      "shape": (None, n_feature_1, n_feature_2), "batch_dim_axis": 0, "time_dim_axis": 1, "feature_dim_axis": 2})


def test_reshape_a_b_F_to_b_aF():
  n_batch, n_time, n_feature_1, n_feature_2 = 3, 5, 8, 6

  def model_func(wrapped_import, inputs: torch.Tensor):
    # test case (..., a, b, F,...) -> (..., b, a*F,...)
    inputs = inputs.transpose(1, 2).contiguous()
    return inputs.view(n_batch, n_time, n_feature_1 * n_feature_2)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_feature_1, n_time, n_feature_2)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x, inputs_data_kwargs={
      "shape": (n_feature_1, None, n_feature_2), "batch_dim_axis": 0, "time_dim_axis": 2, "feature_dim_axis": 3})


def test_reshape_a_b_1_to_a_b():
  n_batch, n_time, n_feature = 2, 7, 1

  def model_func(wrapped_import, inputs: torch.Tensor):
    # test case (a, b, 1) -> (a, b)
    out = inputs.view(inputs.shape[:2])  # (B, T, F)
    return out

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_time, n_feature)).astype("float32")
  verify_torch_and_convert_to_returnn(
    model_func, inputs=x, returnn_dummy_input_shape=x.shape,
    inputs_data_kwargs={
      "shape": (None, n_feature), "batch_dim_axis": 0, "time_dim_axis": 1, "feature_dim_axis": 2})


def test_pad():
  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch.nn.functional as F
    else:
      F = wrapped_import("torch.nn.functional")

    inputs = F.pad(inputs, (1, 1, 2, 2))
    inputs = F.pad(inputs, (1, 1))
    return inputs

  x = numpy.zeros((1, 1, 4, 4)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_pad_time_btf():
  n_batch, n_time, n_feat = 3, 7, 5

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch.nn.functional as F
    else:
      F = wrapped_import("torch.nn.functional")

    inputs = F.pad(inputs, (0, 0, 2, 2))
    return inputs

  x = numpy.ones((n_batch, n_time, n_feat)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)
  verify_torch_and_convert_to_returnn(model_func, inputs=x, inputs_data_kwargs={
    "batch_dim_axis": 0, "time_dim_axis": 1, "feature_dim_axis": 2, "shape": (None, n_feat)})


def test_constant_pad_1d():
  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")

    mod = torch.nn.ConstantPad1d((4, 1), 0)
    return mod(inputs)

  x = numpy.zeros((3, 5, 7)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_functional_conv():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7
  kernel_size = 3

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
      import torch.nn.functional as F
    else:
      torch = wrapped_import("torch")
      F = wrapped_import("torch.nn.functional")
    rnd = numpy.random.RandomState(42)
    weight = rnd.normal(0., 1., (n_out, n_in, kernel_size)).astype("float32")
    bias = rnd.normal(0., 1., (n_out,)).astype("float32")
    weight = torch.from_numpy(weight)
    bias = torch.from_numpy(bias)
    return F.conv1d(
      inputs,
      weight=weight, bias=bias,
      stride=2)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_functional_conv_no_bias():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7
  kernel_size = 3

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
      import torch.nn.functional as F
    else:
      torch = wrapped_import("torch")
      F = wrapped_import("torch.nn.functional")
    rnd = numpy.random.RandomState(42)
    weight = rnd.normal(0., 1., (n_out, n_in, kernel_size)).astype("float32")
    weight = torch.from_numpy(weight)
    return F.conv1d(inputs, weight=weight, stride=2)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_functional_conv2d():
  n_in, n_out = 11, 13
  n_batch, n_time1, n_time2 = 3, 17, 19
  kernel_size = (3, 5)

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
      import torch.nn.functional as F
    else:
      torch = wrapped_import("torch")
      F = wrapped_import("torch.nn.functional")
    rnd = numpy.random.RandomState(42)
    weight = rnd.normal(0., 1., (n_out, n_in) + kernel_size).astype("float32")
    bias = rnd.normal(0., 1., (n_out,)).astype("float32")
    weight = torch.from_numpy(weight)
    bias = torch.from_numpy(bias)
    return F.conv2d(
      inputs,
      weight=weight, bias=bias,
      stride=2)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time1, n_time2)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_functional_conv_transposed():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7
  kernel_size = 3

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
      import torch.nn.functional as F
    else:
      torch = wrapped_import("torch")
      F = wrapped_import("torch.nn.functional")
    rnd = numpy.random.RandomState(42)
    weight = rnd.normal(0., 1., (n_in, n_out, kernel_size)).astype("float32")
    weight = torch.from_numpy(weight)
    return F.conv_transpose1d(inputs, weight=weight, stride=2)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_batch_norm():
  n_in, n_batch, n_time = 11, 3, 7
  for train in [True, False]:

    def model_func(wrapped_import, inputs: torch.Tensor):
      if typing.TYPE_CHECKING or not wrapped_import:
        import torch
      else:
        torch = wrapped_import("torch")
      model = torch.nn.BatchNorm1d(n_in)
      if not train:
        model.eval()
      out = model(inputs)
      if train:
        model.reset_running_stats()  # for the test, such that we start with initial running mean/var
      return out

    x = numpy.ones((n_batch, n_in, n_time)).astype("float32")
    verify_torch_and_convert_to_returnn(model_func, inputs=x, train=train)

    rnd = numpy.random.RandomState(42)
    x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
    verify_torch_and_convert_to_returnn(model_func, inputs=x, train=train)


def test_batch_norm_running_stats():
  from pytorch_to_returnn.torch.nn import Module as ModuleReturnn
  from pytorch_to_returnn.naming import Naming, ModuleEntry
  from torch.nn import Module as ModuleTorch
  n_in, n_batch, n_time = 11, 3, 7
  mean_torch = None
  mean_returnn = None

  def model_func(wrapped_import, inputs: torch.Tensor):
    nonlocal mean_torch, mean_returnn
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    model = torch.nn.BatchNorm1d(n_in)
    out = model(inputs)
    if isinstance(model, ModuleTorch):
      mean_torch = model.running_mean.detach().cpu().numpy().copy()
    elif isinstance(model, ModuleReturnn):
      naming = Naming.get_instance()
      if naming.import_params_from_torch_namespace:  # only then we have the params
        module_entry = naming.modules[model]
        assert isinstance(module_entry, ModuleEntry)
        assert len(module_entry.calls) == 1
        call = module_entry.calls[0]
        assert call.returnn_layer
        mean_returnn = tf.squeeze(call.returnn_layer.params["batch_norm/mean"]).eval()
    model.reset_running_stats()  # for the test, such that we start with initial running mean/var
    return out

  x = numpy.ones((n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x, train=True)
  assert mean_returnn is not None and mean_torch is not None
  print(mean_torch)
  numpy.testing.assert_allclose(mean_torch[0], 0.1, rtol=0, atol=1e-5)  # default momentum 0.1
  numpy.testing.assert_allclose(mean_returnn, mean_torch, rtol=0, atol=1e-5)

  mean_returnn = mean_torch = None
  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x, train=True)
  assert mean_returnn is not None and mean_torch is not None
  print(mean_torch)
  numpy.testing.assert_allclose(mean_returnn, mean_torch, rtol=0, atol=1e-5)


def test_fp32_layer_norm():
  n_in, n_batch, n_time = 11, 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
      import torch.nn.functional as F
    else:
      torch = wrapped_import("torch")
      F = wrapped_import("torch.nn.functional")

    # copy of Fp32LayerNorm from fairseq
    class Fp32LayerNorm(torch.nn.LayerNorm):
      def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

      def forward(self, input):
        output = F.layer_norm(
          input.float(),
          self.normalized_shape,
          self.weight.float() if self.weight is not None else None,
          self.bias.float() if self.bias is not None else None,
          self.eps
        )
        return output.type_as(input)

    model = Fp32LayerNorm(n_in, elementwise_affine=True)
    out = inputs.transpose(-2, -1)
    out = model(out)
    out = out.transpose(-2, -1)
    return out

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x, inputs_data_kwargs={
    "batch_dim_axis": 0, "time_dim_axis": 2, "feature_dim_axis": 1, "shape": (n_in, None)})


def test_group_norm():
  n_batch, n_time = 4, 20
  for num_groups, num_channels in [(1, 5), (5, 5)]:

    def model_func(wrapped_import, inputs: torch.Tensor):
      if typing.TYPE_CHECKING or not wrapped_import:
        import torch
      else:
        torch = wrapped_import("torch")

      model = torch.nn.GroupNorm(num_groups, num_channels)
      out = model(inputs)
      return out

    print(f"test for num_groups={num_groups}, num_channels={num_channels}")
    rnd = numpy.random.RandomState(42)
    x = rnd.normal(0., 1., (n_batch, num_channels, n_time)).astype("float32")
    verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_fp32_group_norm():
  n_batch, n_time = 3, 17
  for num_groups, num_channels in [(1, 5), (5, 5)]:

    def model_func(wrapped_import, inputs: torch.Tensor):
      if typing.TYPE_CHECKING or not wrapped_import:
        import torch
        import torch.nn.functional as F
      else:
        torch = wrapped_import("torch")
        F = wrapped_import("torch.nn.functional")

      # copy of Fp32GroupNorm from fairseq
      class Fp32GroupNorm(torch.nn.GroupNorm):
        def __init__(self, *args, **kwargs):
          super().__init__(*args, **kwargs)

        def forward(self, input):
          output = F.group_norm(
            input.float(),
            self.num_groups,
            self.weight.float() if self.weight is not None else None,
            self.bias.float() if self.bias is not None else None,
            self.eps)
          return output.type_as(input)

      model = Fp32GroupNorm(num_groups, num_channels)
      out = model(inputs)
      return out

    print(f"test for num_groups={num_groups}, num_channels={num_channels}")
    rnd = numpy.random.RandomState(42)
    x = rnd.normal(0., 1., (n_batch, num_channels, n_time)).astype("float32")
    verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_fp32_group_norm_subnetwork():
  n_batch, n_time = 3, 17
  for num_groups, num_channels in [(1, 5), (5, 5)]:

    def model_func(wrapped_import, inputs: torch.Tensor):
      if typing.TYPE_CHECKING or not wrapped_import:
        import torch
        import torch.nn.functional as F
      else:
        torch = wrapped_import("torch")
        F = wrapped_import("torch.nn.functional")

      # copy of Fp32GroupNorm from fairseq
      class Fp32GroupNorm(torch.nn.GroupNorm):
        def __init__(self, *args, **kwargs):
          super().__init__(*args, **kwargs)

        def forward(self, input):
          output = F.group_norm(
            input.float(),
            self.num_groups,
            self.weight.float() if self.weight is not None else None,
            self.bias.float() if self.bias is not None else None,
            self.eps)
          return output.type_as(input)

      model = torch.nn.Sequential(Fp32GroupNorm(num_groups, num_channels))
      out = model(inputs)
      return out

    print(f"test for num_groups={num_groups}, num_channels={num_channels}")
    rnd = numpy.random.RandomState(42)
    x = rnd.normal(0., 1., (n_batch, num_channels, n_time)).astype("float32")
    verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_multi_head_attention_forward():
  n_batch, n_time, n_feature = 3, 17, 8

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
      import torch.nn.functional as F
    else:
      torch = wrapped_import("torch")
      F = wrapped_import("torch.nn.functional")

    class MultiHeadAttention(torch.nn.Module):
      def __init__(self, embed_dim, num_heads,):
        super(MultiHeadAttention, self).__init__()

        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.dummy_weight = torch.ones(embed_dim, embed_dim)
        self.dummy_in_bias = torch.zeros(3 * embed_dim)
        self.dummy_out_bias = torch.zeros(embed_dim)

        self.k_proj = torch.nn.Linear(embed_dim, embed_dim, bias=False)
        self.q_proj = torch.nn.Linear(embed_dim, embed_dim, bias=False)
        self.v_proj = torch.nn.Linear(embed_dim, embed_dim, bias=False)
        torch.nn.init.xavier_uniform_(self.k_proj.weight)
        torch.nn.init.xavier_uniform_(self.v_proj.weight)
        torch.nn.init.xavier_uniform_(self.q_proj.weight)

      def forward(self, query, key, value):
        attn_output, attn_output_weights = F.multi_head_attention_forward(
          query, key, value, self.embed_dim, self.num_heads,
          in_proj_weight=self.dummy_weight, in_proj_bias=self.dummy_in_bias,
          bias_k=None, bias_v=None, add_zero_attn=False, dropout_p=0.0,
          out_proj_weight=self.dummy_weight, out_proj_bias=self.dummy_out_bias,
          need_weights=False, use_separate_proj_weight=True,
          k_proj_weight=self.k_proj.weight, q_proj_weight=self.q_proj.weight, v_proj_weight=self.v_proj.weight)
        return attn_output


    model = MultiHeadAttention(embed_dim=n_feature, num_heads=2)
    inputs = inputs
    out = model(query=inputs, key=inputs, value=inputs)
    return out

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_time, n_batch, n_feature)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x, inputs_data_kwargs={
    "time_dim_axis": 0, "batch_dim_axis": 1, "feature_dim_axis": 2, "shape": (None, n_feature)})


def test_unsqueeze():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    return inputs.unsqueeze(2)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_unsqueeze2():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    return inputs.unsqueeze(3)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_broadcast_with_different_axes_types():
  n_batch, n_time, n_feature = 3, 7, 5

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    weight = torch.ones(n_feature)
    return inputs * weight.view(1, -1, 1)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_feature, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_broadcast_after_transpose():
  n_batch, n_time, n_feature = 3, 7, 5

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    weight = torch.ones(n_feature)
    return inputs.transpose(1, 2) * weight

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_feature, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_broadcast_add_bias():
  n_batch, n_feature = 3, 5

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    rnd = numpy.random.RandomState(42)
    bias = rnd.normal(0., 1., (n_feature,)).astype("float32")
    bias = torch.from_numpy(bias)
    return inputs + bias

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_feature)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x, inputs_data_kwargs={"time_dim_axis": None})


def test_movedim():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    x = inputs  # (B,F,T)
    x = torch.nn.Conv1d(n_in, n_out, 3)(x)  # make it (B,T,F) in RETURNN. (B,F,T) in Torch.
    x = torch.movedim(x, 0, 1)  # stay (B,T,F) in RETURNN. (F,B,T) in Torch.
    return x

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_movedim_2():
  n_batch, n_time, n_feature = 3, 5, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    out = torch.movedim(inputs, 0, 2)
    return out

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_feature, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_permute():
  n_batch, n_time, n_feature = 3, 5, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    out = inputs.permute(2, 0, 1)
    return out

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_feature, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_flatten_batch():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    x = inputs  # (B,F,T)
    x = torch.movedim(x, 1, 0)  # (F,B,T)
    x = torch.reshape(x, (x.shape[0], -1))  # (F,B*T)
    return x

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  converter = verify_torch_and_convert_to_returnn(model_func, inputs=x)
  assert converter.returnn_net_dict["Flatten"]["class"] == "flatten_batch", pformat(converter.returnn_net_dict)


def test_reduce_sum():
  n_batch, n_time, n_feature = 3, 5, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    out = inputs.sum(dim=1)
    return out

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_feature, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_slice_1d():
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    x = inputs[0]  # (T)
    x = x[:3]
    return x

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_slice_2d():
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    x = inputs[:, 3:]
    return x

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_expand():
  n_batch, n_feat = 5, 1

  def model_func(wrapped_import, inputs: torch.Tensor):
    x = inputs.expand(3, 2, -1, 7)
    return x

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_feat)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_arange():
  n_batch, n_feat = 1, 5

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    arange = torch.arange(inputs.shape[1])
    return inputs + torch.reshape(arange, (1, -1))

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_feat)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_const_with_batch_and_gather():
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    batch_size = inputs.shape[0]
    x = inputs.new_zeros((2, batch_size, 13))  # const, shape (2,B,13). in RETURNN (B,2,13)
    x = x[0]  # shape (B,13). in RETURNN (B,13).
    x = x + inputs[:, :1]  # (B,13). just do this to have dep on inputs
    return x

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_lstm_2l():
  n_batch, n_time = 3, 7
  n_in, n_out = 11, 13
  n_layers = 2

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    batch_size = inputs.shape[1]
    hidden = inputs.new_zeros((n_layers, batch_size, n_out))  # const, shape (2,B,13). in RETURNN (B,2,13)
    hidden = (hidden, hidden)
    lstm = torch.nn.LSTM(n_in, n_out, n_layers)
    out, hidden_ = lstm(inputs, hidden)
    return out

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_time, n_batch, n_in)).astype("float32")
  verify_torch_and_convert_to_returnn(
    model_func, inputs=x,
    inputs_data_kwargs={"shape": (None, n_in), "batch_dim_axis": 1})


def test_blstm():
  n_batch, n_time = 3, 7
  n_in, n_out = 11, 13
  n_layers = 1

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    batch_size = inputs.shape[1]
    hidden = inputs.new_zeros((2 * n_layers, batch_size, n_out))
    hidden = (hidden, hidden)
    blstm = torch.nn.LSTM(n_in, n_out, n_layers, bidirectional=True)
    out, hidden_ = blstm(inputs, hidden)
    return out

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_time, n_batch, n_in)).astype("float32")
  verify_torch_and_convert_to_returnn(
    model_func, inputs=x,
    inputs_data_kwargs={"shape": (None, n_in), "batch_dim_axis": 1})


def test_multiple_outputs():
  n_batch, n_time = 3, 7
  n_in, n_out = 11, 13
  n_layers = 1

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")

    class InnerModel(torch.nn.Module):
      def __init__(self):
        super(InnerModel, self).__init__()
        self.lstm = torch.nn.LSTM(n_in, n_out, n_layers)

      def forward(self, x, hidden):
        return self.lstm(x, hidden)

    # Wrap InnerModel in another model, to make sure InnerModel is not on the flattened root namespace.
    class OuterModel(torch.nn.Module):
      def __init__(self):
        super(OuterModel, self).__init__()
        self.model = InnerModel()

      def forward(self, input):
        batch_size = input.shape[1]
        hidden = torch.zeros((n_layers, batch_size, n_out))  # const, shape (1,B,13). in RETURNN (B,1,13)
        hidden = (hidden, hidden)
        out, (h, c) = self.model(input, hidden)  # out is (T,B,out), h/c is (1,B,out)
        return out[0] + h[0] + c[0]  # combine all, to set dependencies. (B,out)

    model = OuterModel()
    return model(inputs)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_time, n_batch, n_in)).astype("float32")
  verify_torch_and_convert_to_returnn(
    model_func, inputs=x,
    inputs_data_kwargs={"shape": (None, n_in), "batch_dim_axis": 1})


def test_forward_with_kwargs():
  n_batch, n_time, n_feature = 3, 7, 5

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")

    class MyModel(torch.nn.Module):
      def __init__(self, bias=0):
        super(MyModel, self).__init__()
        self.bias = bias

      def forward(self, x, *, add_bias=None):
        assert isinstance(add_bias, bool)
        if add_bias:
          x += self.bias
        return x

    model = MyModel(bias=1)
    out = model(inputs, add_bias=True)
    return out

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_feature, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_gated_linear_unit_via_chunk():
  n_in, n_out = 12, 13
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")

    class GLU(torch.nn.Module):
      def __init__(self, dim):
        super().__init__()
        self.dim = dim

      def forward(self, x):
        out, gate = x.chunk(2, dim=self.dim)
        return out * gate.sigmoid()

    model = GLU(dim=1)
    return model(inputs)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_depth_wise_conv1d():
  n_in, n_out = 12, 24
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
      import torch.nn.functional as F
    else:
      torch = wrapped_import("torch")
      F = wrapped_import("torch.nn.functional")

    class DepthWiseConv1d(torch.nn.Module):
      def __init__(self, chan_in, chan_out, kernel_size, padding):
        super().__init__()
        self.padding = padding
        self.conv = torch.nn.Conv1d(chan_in, chan_out, kernel_size, groups=chan_in)

      def forward(self, x):
        x = F.pad(x, self.padding)
        return self.conv(x)

    model = DepthWiseConv1d(n_in, n_out, [3], (0, 0))
    return model(inputs)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_dummy_input_shape():
  n_in, n_out = 11, 11
  n_batch, n_time = 3, 999

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    model = torch.nn.Conv1d(in_channels=n_in, out_channels=n_out, kernel_size=5, stride=5)
    x = inputs
    for i in range(3):
      x = model(x)
    return x

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x, returnn_dummy_input_shape=x.shape)


if __name__ == "__main__":
  if len(sys.argv) <= 1:
    for k, v in sorted(globals().items()):
      if k.startswith("test_"):
        print("-" * 40)
        print("Executing: %s" % k)
        try:
          v()
        except unittest.SkipTest as exc:
          print("SkipTest:", exc)
        print("-" * 40)
    print("Finished all tests.")
  else:
    assert len(sys.argv) >= 2
    for arg in sys.argv[1:]:
      print("Executing: %s" % arg)
      if arg in globals():
        globals()[arg]()  # assume function and execute
      else:
        eval(arg)  # assume Python code and execute
